#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Module containing class methods to write within xml"""
# Libraries
from lxml import etree

import datamodel


class GenerateXML:
    """Class to generate XML"""
    def __init__(self, xml_file):
        """Initialize XML structure/tags and file's object"""
        self.root = etree.Element("systemAnalysis")
        fun_arch = etree.SubElement(self.root, "funcArch")
        fun_arch_tags = ['functionList', 'dataList', 'stateList', 'transitionList',
                         'functionalElementList', 'functionalInterfaceList']
        for tag in fun_arch_tags:
            etree.SubElement(fun_arch, tag)
        phy_arch = etree.SubElement(self.root, "phyArch")
        phy_arch_tags = ['physicalElementList', 'physicalInterfaceList']
        for tag in phy_arch_tags:
            etree.SubElement(phy_arch, tag)
        viewpoint = etree.SubElement(self.root, "viewPoint")
        viewpoint_tags = ['viewList', 'attributeList', 'typeList']
        for tag in viewpoint_tags:
            etree.SubElement(viewpoint, tag)
        self.tree = etree.ElementTree(self.root)

        if len(xml_file) > 0:
            self.file = xml_file
        else:
            self.file = "Output.xml"

    def write_function(self, function_list):
        """Method to write functions from function's list"""
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            if root.find('.//functionList') is None:
                etree.SubElement(root, 'functionList')
            # Loop on each flow/data
            for function_list_tag in root.findall(".//functionList"):
                # Loop on function's list
                for function in function_list:
                    function_tag = etree.SubElement(function_list_tag, "function",
                                                    {'id': function.id,
                                                     'name': function.name,
                                                     'type': str(function.type),
                                                     'alias': function.alias,
                                                     'derived': function.derived})

                    _functional_part_list_tag = etree.SubElement(function_tag, "functionPartList")

        self.write()

    def write_data(self, data_list):
        """Method to add data flows"""
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            if root.find('.//dataList') is None:
                etree.SubElement(root, 'dataList')
            for data_list_tag in root.findall('.//dataList'):
                for data in data_list:
                    existing_data_tag = data_list_tag.find('.//dataList/data')
                    if existing_data_tag is not None:
                        tag = existing_data_tag
                    else:
                        tag = data_list_tag
                    data_tag = etree.SubElement(tag, "data", {'name': data.name,
                                                              'type': str(data.type),
                                                              'id': data.id})

                    _consumer_list_tag = etree.SubElement(data_tag, "consumerList")

                    _producer_list_tag = etree.SubElement(data_tag, "producerList")

                    _predecessor_list_tag = etree.SubElement(data_tag, "predecessorList")

        self.write()

    def write_consumer(self, consumer_list):
        """Method to write consumers by list [data_name, function]"""
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            for i in consumer_list:
                for consumer_list_tag in root.findall(".//dataList/data[@name='" + str(i[0])
                                                      + "']/consumerList"):
                    if not i[1].operand:
                        _consumer_tag = etree.SubElement(consumer_list_tag, "consumer",
                                                         {'id': i[1].id, 'role': "none"})
                    else:
                        _consumer_tag = etree.SubElement(consumer_list_tag, "consumer",
                                                         {'id': i[1].id, 'role': i[1].operand})

        self.write()

    def write_producer(self, producer_list):
        """Method to write producers by list [data_name, function]"""
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            for i in producer_list:
                for producer_list_tag in root.findall(".//dataList/data[@name='" + str(i[0])
                                                      + "']/producerList"):

                    _producer_tag = etree.SubElement(producer_list_tag, "producer", {'id': i[1].id})

        self.write()

    def write_predecessor(self, predecessor_list):
        """Method to write predecessors by list [data, predecessor]"""
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            for i in predecessor_list:
                for predecessor_list_tag in root.findall(
                        ".//dataList/data[@name='" + str(i[0].name) + "']/predecessorList"):

                    _producer_tag = etree.SubElement(predecessor_list_tag, "predecessor",
                                                     {'id': i[1].id})

        self.write()

    def delete_single_consumer_producer(self, data, function, value):
        """Method to delete the parents (consumer or producer) when flow is within a component"""
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            # Find specific consumer/producer to delete
            for tag in root.findall(
                    ".//data[@name='" + data + "']/" + value + "List/" + value + "[@id='"
                    + function.id + "']"):
                tag.getparent().remove(tag)
        self.write()

    def write(self):
        """Method to write within XML file"""
        with open(self.file, "wb") as file:
            self.tree.write(file, encoding='utf-8', xml_declaration=True, pretty_print=True)

    def write_state(self, state_list):
        """Method to write states from state's list"""
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            if root.find('.//stateList') is None:
                etree.SubElement(root, 'stateList')
            for state_list_tag in root.findall(".//stateList"):
                for state in state_list:
                    state_tag = etree.SubElement(state_list_tag, "state",
                                                 {'id': state.id, 'name': state.name,
                                                  'type': str(state.type), 'alias': state.alias})

                    _state_part_list_tag = etree.SubElement(state_tag, "statePartList")
                    _allocated_function_list_tag = etree.SubElement(state_tag,
                                                                    "allocatedFunctionList")
        self.write()

    def write_transition(self, transition_list):
        """Method to write transition from transition's list"""
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            if root.find('.//transitionList') is None:
                etree.SubElement(root, 'transitionList')
            for transition_list_tag in root.findall(".//transitionList"):
                for transition in transition_list:
                    transition_tag = etree.SubElement(transition_list_tag, "transition",
                                                      {'id': transition.id,
                                                       'name': transition.name,
                                                       'type': str(transition.type),
                                                       'alias': transition.alias,
                                                       'source': str(transition.source),
                                                       'destination': str(transition.destination)})
                    _transition_part_list_tag = etree.SubElement(transition_tag, "conditionList")
        self.write()

    def write_transition_condition(self, transition_condition_list):
        """Method to write transition's condition by list [transition, condition]"""
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            for transition in root.findall(".//transition"):
                for tra, condition in transition_condition_list:
                    if transition.get('id') == tra.id:
                        tag = transition.find('conditionList')
                        _state_part_tag = etree.SubElement(tag, "condition",
                                                           {'text': str(condition)})
        self.write()

    def write_source(self, transition_source_list):
        """Method to write transition's source by list [transition, source]"""
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            for transition_src in transition_source_list:
                for state_tag in root.findall(".//transition[@id='" + transition_src[0].id + "']"):
                    state_tag.set('source', transition_src[1].id)
        self.write()

    def write_destination(self, transition_destination_list):
        """Method to write transition's destination by list [transition, destination]"""
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            for transition_dest in transition_destination_list:
                for state_tag in root.findall(".//transition[@id='" + transition_dest[0].id + "']"):
                    state_tag.set('destination', transition_dest[1].id)
        self.write()

    def write_functional_element(self, functional_element_list):
        """Method to write functional element from functional elements list"""
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            if root.find('.//functionalElementList') is None:
                etree.SubElement(root, 'functionalElementList')
            for functional_element_list_tag in root.findall(".//functionalElementList"):
                for functional_element in functional_element_list:
                    functional_element_tag = etree.SubElement(
                        functional_element_list_tag, "functionalElement",
                        {
                            'id': functional_element.id,
                            'name': functional_element.name,
                            'type': str(
                                functional_element.type),
                            'alias': functional_element.alias,
                            'derived': functional_element.derived}
                    )

                    _fun_elem_part_list_tag = etree.SubElement(functional_element_tag,
                                                               "functionalElementPartList")
                    _allocated_state_list_tag = etree.SubElement(functional_element_tag,
                                                                 "allocatedStateList")
                    _allocated_function_list_tag = etree.SubElement(functional_element_tag,
                                                                    "allocatedFunctionList")
                    _exposed_interface_list_tag = etree.SubElement(functional_element_tag,
                                                                   "exposedInterfaceList")
        self.write()

    def write_exposed_interface(self, fun_elem_inter_list):
        """Method to write exposed interfaces by list [fun_elem/phy_elem, exposed_interface]"""
        if isinstance(fun_elem_inter_list[0][0], datamodel.FunctionalElement):
            string_tag = ".//functionalElement"
        else:
            string_tag = ".//physicalElement"
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            for functional_element in root.findall(string_tag):
                if functional_element.find('exposedInterfaceList') is None:
                    etree.SubElement(functional_element, 'exposedInterfaceList')

                for fun_elem, inter in fun_elem_inter_list:
                    if functional_element.get('id') == fun_elem.id:
                        tag = functional_element.find('exposedInterfaceList')
                        _exposed_interface_tag = etree.SubElement(tag, "exposedInterface",
                                                                  {'id': inter.id})
        self.write()

    def write_view(self, view_list):
        """Method to write views from view's list"""
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            if root.find('.//viewList') is None:
                etree.SubElement(root, 'viewList')
            for view_list_tag in root.findall(".//viewList"):
                for view in view_list:
                    view_tag = etree.SubElement(view_list_tag, "view",
                                                 {'id': view.id, 'name': view.name,
                                                  'type': str(view.type)})
                    _allocated_item_list_tag = etree.SubElement(view_tag, "allocatedItemList")
        self.write()

    def write_attribute(self, attribute_list):
        """Method to write attributes from attribute's list"""
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            if root.find('.//attributeList') is None:
                etree.SubElement(root, 'attributeList')
            for attribute_list_tag in root.findall(".//attributeList"):
                for attribute in attribute_list:
                    attribute_tag = etree.SubElement(attribute_list_tag, "attribute",
                                                     {'id': attribute.id, 'name': attribute.name,
                                                      'type': str(attribute.type),
                                                      'alias': attribute.alias})

                    _described_item_list_tag = etree.SubElement(attribute_tag, "describedItemList")

        self.write()

    def write_described_attribute_item(self, attribute_item_list):
        """Method to write described item by list [attribute, (described_item, value)]"""
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            # Loop on each attribute
            for attribute_element in root.findall(".//attribute"):
                for attribute, item in attribute_item_list:
                    if attribute_element.get('id') == attribute.id:
                        tag = attribute_element.find('describedItemList')
                        _allocated_item_tag = etree.SubElement(tag, "describedItem",
                                                               {'id': item[0].id, 'value': item[1]})
        self.write()

    def write_functional_interface(self, functional_interface_list):
        """Method to write functional interfaces from interface's list"""
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            if root.find('.//functionalInterfaceList') is None:
                etree.SubElement(root, 'functionalInterfaceList')
            for fun_interface_list_tag in root.findall(".//functionalInterfaceList"):
                for fun_interface in functional_interface_list:
                    fun_interface_tag = etree.SubElement(fun_interface_list_tag,
                                                         "functionalInterface",
                                                         {'id': fun_interface.id,
                                                          'name': fun_interface.name,
                                                          'type': str(fun_interface.type),
                                                          'alias': fun_interface.alias,
                                                          'derived': fun_interface.derived})
                    _allocated_data_list_tag = etree.SubElement(fun_interface_tag,
                                                                "allocatedDataList")
        self.write()

    def write_physical_element(self, physical_element_list):
        """Method to write physical element from physical elements list"""
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            if root.find('.//physicalElementList') is None:
                etree.SubElement(root, 'physicalElementList')
            for physical_element_list_tag in root.findall(".//physicalElementList"):
                for physical_element in physical_element_list:
                    physical_element_tag = etree.SubElement(
                        physical_element_list_tag, "physicalElement",
                        {
                            'id': physical_element.id,
                            'name': physical_element.name,
                            'type': str(
                                physical_element.type),
                            'alias': physical_element.alias,
                            'derived': physical_element.derived}
                    )

                    _phy_elem_part_list_tag = etree.SubElement(physical_element_tag,
                                                               "physicalElementPartList")
                    _allocated_fun_elem_list_tag = etree.SubElement(
                        physical_element_tag, "allocatedFunctionalElementList")
                    _exposed_interface_list_tag = etree.SubElement(physical_element_tag,
                                                                   "exposedInterfaceList")
        self.write()

    def write_physical_interface(self, physical_interface_list):
        """Method to write physical interfaces from interface's list"""
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            if root.find('.//physicalInterfaceList') is None:
                etree.SubElement(root, 'physicalInterfaceList')
            for phy_interface_list_tag in root.findall(".//physicalInterfaceList"):
                for phy_interface in physical_interface_list:
                    phy_interface_tag = etree.SubElement(phy_interface_list_tag,
                                                         "physicalInterface",
                                                         {'id': phy_interface.id,
                                                          'name': phy_interface.name,
                                                          'type': str(phy_interface.type),
                                                          'alias': phy_interface.alias,
                                                          'derived': phy_interface.derived})
                    _allocated_fun_inter_list_tag = etree.SubElement(
                        phy_interface_tag, "allocatedFunctionalInterfaceList")
        self.write()

    def write_object_alias(self, object_list):
        """Method to write object's alias by list [object]"""
        elem_tag = get_object_tag(object_list[0])
        if elem_tag:
            with open(self.file, 'rb') as file:
                parser = etree.XMLParser(remove_blank_text=True)
                root = self.tree.parse(file, parser)
                for obj in object_list:
                    for obj_tag in root.findall(".//" + elem_tag + "[@id='" + obj.id + "']"):
                        obj_tag.set('alias', str(obj.alias))
            self.write()

    def write_derived(self, derived_list):
        """Method to write derived by list [Object]"""
        for obj in derived_list:
            elem_tag = get_object_tag(obj)
            if elem_tag in derived_obj_tag:
                with open(self.file, 'rb') as file:
                    parser = etree.XMLParser(remove_blank_text=True)
                    root = self.tree.parse(file, parser)
                    for elem in derived_list:
                        for fun_inter_tag in root.findall(".//" + elem_tag + "[@id='" +
                                                          elem.id + "']"):
                            fun_inter_tag.set('derived', str(elem.derived.id))
            self.write()

    def write_object_type(self, object_list):
        """Method to write object's type by list [object]"""
        elem_tag = get_object_tag(object_list[0])
        if elem_tag:
            with open(self.file, 'rb') as file:
                parser = etree.XMLParser(remove_blank_text=True)
                root = self.tree.parse(file, parser)
                for obj in object_list:
                    for object_tag in root.findall(".//" + elem_tag + "[@id='" + obj.id + "']"):
                        object_tag.set('type', obj.type)
            self.write()

    def write_object_child(self, object_child_list):
        """Method to write child by list [parent, child]"""
        elem_tag = get_object_tag(object_child_list[0][0])
        if elem_tag:
            with open(self.file, 'rb') as file:
                parser = etree.XMLParser(remove_blank_text=True)
                root = self.tree.parse(file, parser)
                for obj in root.findall(".//" + elem_tag):
                    for parent, child in object_child_list:
                        if obj.get('id') == parent.id:
                            tag = obj.find(elem_tag + 'PartList')
                            _obj_element_part_tag = etree.SubElement(tag,
                                                                     elem_tag + 'Part',
                                                                     {'id': child.id})
            self.write()

    def delete_object(self, object_list):
        """Method to delete objects by list [object]"""
        elem_tag = get_object_tag(object_list[0])
        if elem_tag:
            with open(self.file, 'rb') as file:
                parser = etree.XMLParser(remove_blank_text=True)
                root = self.tree.parse(file, parser)
                for obj in object_list:
                    for obj_tag in root.findall(".//" + elem_tag + "[@id='" + obj.id + "']"):
                        obj_tag.getparent().remove(obj_tag)
            self.write()

    def write_objects_allocation(self, objects_list):
        """Method to write allocated objects from list [Object, Object]"""
        elem_tag = get_object_tag(objects_list[0][0])
        if elem_tag:
            with open(self.file, 'rb') as file:
                parser = etree.XMLParser(remove_blank_text=True)
                root = self.tree.parse(file, parser)
                for obj_tag in root.findall(".//" + elem_tag):
                    for obj, obj_to_alloc in objects_list:
                        if obj_tag.get('id') == obj.id:
                            if elem_tag == "view":
                                alloc_tag = get_allocation_tag(obj)
                            else:
                                alloc_tag = get_allocation_tag(obj_to_alloc)
                            tag = obj_tag.find(alloc_tag + 'List')
                            _allocated_obj_tag = etree.SubElement(
                                tag, alloc_tag, {'id': str(obj_to_alloc.id)})
            self.write()

    def write_type_element(self, type_list):
        """Method to write type element from type's list"""
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            if root.find('.//typeList') is None:
                etree.SubElement(root.find('./viewPoint'), 'typeList')
            for type_list_tag in root.findall(".//typeList"):
                for type_elem in type_list:
                    if isinstance(type_elem.base, datamodel.Type):
                        base_type = type_elem.base.name
                    else:
                        base_type = type_elem.base
                    elem_tag = etree.SubElement(type_list_tag, "type",
                                                {'id': type_elem.id, 'name': type_elem.name,
                                                 'alias': type_elem.alias,
                                                 'base': str(base_type)})

                    # _described_item_list_tag = etree.SubElement(elem_tag, "describedItemList")

        self.write()


derived_obj_tag = ("physicalInterface",
                   "physicalElement",
                   "function",
                   "functionalElement",
                   "functionalInterface")


def get_object_tag(wanted_object):
    """Get xml element tag corresponding to wanted_object"""
    elem_tag = None
    if isinstance(wanted_object, datamodel.PhysicalInterface):
        elem_tag = "physicalInterface"
    elif isinstance(wanted_object, datamodel.PhysicalElement):
        elem_tag = "physicalElement"
    elif isinstance(wanted_object, datamodel.Function):
        elem_tag = "function"
    elif isinstance(wanted_object, datamodel.FunctionalElement):
        elem_tag = "functionalElement"
    elif isinstance(wanted_object, datamodel.FunctionalInterface):
        elem_tag = "functionalInterface"
    elif isinstance(wanted_object, datamodel.Attribute):
        elem_tag = "attribute"
    elif isinstance(wanted_object, datamodel.Transition):
        elem_tag = "transition"
    elif isinstance(wanted_object, datamodel.State):
        elem_tag = "state"
    elif isinstance(wanted_object, datamodel.Data):
        elem_tag = "data"
    elif isinstance(wanted_object, datamodel.View):
        elem_tag = "view"
    elif isinstance(wanted_object, datamodel.Type):
        elem_tag = "type"
    return elem_tag


def get_allocation_tag(wanted_object):
    """Get xml allocation tag corresponding to wanted_object"""
    elem_tag = None
    if isinstance(wanted_object, datamodel.Function):
        elem_tag = "allocatedFunction"
    elif isinstance(wanted_object, datamodel.FunctionalElement):
        elem_tag = "allocatedFunctionalElement"
    elif isinstance(wanted_object, datamodel.FunctionalInterface):
        elem_tag = "allocatedFunctionalInterface"
    elif isinstance(wanted_object, datamodel.State):
        elem_tag = "allocatedState"
    elif isinstance(wanted_object, datamodel.Data):
        elem_tag = "allocatedData"
    elif isinstance(wanted_object, datamodel.View):
        elem_tag = "allocatedItem"
    return elem_tag



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Module with methods relative to Viewpoint section"""
# Libraries
import uuid
import re

import datamodel
from . import shared_orchestrator
from .question_answer import get_objects_names, check_get_object


def add_view(view_name_str, xml_view_list, output_xml):
    """
        Check if each string in view_name_str is not already corresponding to an actual
        object's name, create new View() object, instantiate it, write it within XML and
        then returns update_list.

            Parameters:
                view_name_str ([str]) : Lists of string from jarvis cell
                xml_view_list ([Function]) : view list from xml parsing
                output_xml (GenerateXML object) : XML's file object

            Returns:
                1 if update, else 0
        """
    view_list = []
    # Create a list with all view names already in the xml
    xml_view_name_list = get_objects_names(xml_view_list)
    # Loop on the list and create set for functions
    if view_name_str not in xml_view_name_list:
        # Instantiate view class
        view = datamodel.View()
        # Set view's name
        view.set_name(str(view_name_str))
        # Set view's type
        view.set_type(datamodel.ViewType.UNKNOWN)
        # Generate and set unique identifier of length 10 integers
        identifier = uuid.uuid4()
        view.set_id(str(identifier.int)[:10])
        # Add view to new set() and existing et() from xml
        xml_view_list.add(view)
        view_list.append(view)
        activate_view(view.name, xml_view_list)
    elif view_name_str in xml_view_name_list:
        # print(view_name + " already exists (not added)")
        activate_view(view_name_str, xml_view_list)

    if not view_list:
        return 0

    output_xml.write_view(view_list)
    for view in view_list:
        print(view.name + " is a View")
    return 1


def activate_view(view_name, xml_view_list):
    """Activates View from view's name str"""
    for view in xml_view_list:
        if view_name == view.name:
            view.set_activation(True)
        else:
            view.set_activation(False)


def check_get_consider(consider_str_list, xml_function_list, xml_fun_elem_list, xml_data_list,
                       xml_view_list, output_xml):
    """
    Check and get all "consider xxx" strings. If corresponds to an actual object not yet added to
    the current view => add it to View object and as allocatedItem within xml
    Args:
        consider_str_list ([strings]): list of strings (separated by comma is possible)
        xml_function_list ([Function]) : Function list from xml parsing
        xml_fun_elem_list ([Fun Elem]) : Functional Element list from xml parsing
        xml_data_list ([Data]) : Data list from xml parsing
        xml_view_list ([View]) : View list from xml parsing
        output_xml (GenerateXML object) : XML's file object

    Returns:
        update ([0/1]) : 1 if update, else 0
    """
    allocated_item_list = []
    # Create lists with all object names/aliases already in the xml
    xml_fun_elem_name_list = get_objects_names(xml_fun_elem_list)
    xml_function_name_list = get_objects_names(xml_function_list)
    xml_data_name_list = get_objects_names(xml_data_list)

    consider_str_list = split_chain_from_string(consider_str_list)

    for consider_str in consider_str_list:
        if consider_str not in [*xml_fun_elem_name_list, *xml_function_name_list,
                                *xml_data_name_list]:
            print(f"Object {consider_str} does not exist, available object types are : "
                  f"Functional Element, Function and Data")
        else:
            result_function = any(item == consider_str for item in xml_function_name_list)
            result_fun_elem = any(item == consider_str for item in xml_fun_elem_name_list)
            result_data = any(item == consider_str for item in xml_data_name_list)

            if result_function:
                allocated_fun = shared_orchestrator.check_add_allocated_item(
                    consider_str, xml_function_list, xml_view_list)
                if allocated_fun:
                    allocated_item_list.append(allocated_fun)
            elif result_fun_elem:
                allocated_fun_elem = shared_orchestrator.check_add_allocated_item(
                    consider_str, xml_fun_elem_list, xml_view_list)
                if allocated_fun_elem:
                    allocated_item_list.append(allocated_fun_elem)
            elif result_data:
                allocated_data = shared_orchestrator.check_add_allocated_item(
                    consider_str, xml_data_list, xml_view_list)
                if allocated_data:
                    allocated_item_list.append(allocated_data)

    update = shared_orchestrator.add_allocation({5: allocated_item_list}, output_xml)

    return update


def split_chain_from_string(consider_str_list):
    """Creates flatten list : from ["A", "B, C, D"] to ["A", "B", "C", "D"]"""
    output_list = []
    for consider_str in consider_str_list:
        if "," in consider_str:
            clean_diagram_object_str = consider_str.replace(" ", "")
            output_list += re.split(r',(?![^[]*\])', clean_diagram_object_str)
        else:
            output_list.append(consider_str)

    return output_list


def add_attribute(attribute_str_list, xml_attribute_list, output_xml):
    """
    Check if each string in xml_attribute_list is not already corresponding to an actual object's
    name/alias, create new Attribute() object, instantiate it, write it within XML and then returns
    update_list.

        Parameters:
            attribute_str_list ([str]) : Lists of string from jarvis cell
            xml_attribute_list ([Attribute]) : Attribute list from xml parsing
            output_xml (GenerateXML object) : XML's file object

        Returns:
            1 if update, else 0
    """
    new_attribute_list = []
    # Create attribute names list already in xml
    xml_attribute_name_list = get_objects_names(xml_attribute_list)
    # Filter attribute_list, keeping only the the ones not already in the xml
    for attribute_name in attribute_str_list:
        if attribute_name not in xml_attribute_name_list:
            new_attribute = datamodel.Attribute()
            new_attribute.set_name(str(attribute_name))
            # Generate and set unique identifier of length 10 integers
            identifier = uuid.uuid4()
            new_attribute.set_id(str(identifier.int)[:10])
            # Not needed, by default unknown
            # new_data.set_type(datamodel.DataType.UNKNOWN)
            # alias is 'none' by default
            new_attribute_list.append(new_attribute)

    if not new_attribute_list:
        return 0

    output_xml.write_attribute(new_attribute_list)
    for attribute in new_attribute_list:
        xml_attribute_list.add(attribute)
        print(attribute.name + " is an attribute")
    return 1


def check_add_object_attribute(described_attribute_list, xml_attribute_list, xml_function_list,
                               xml_fun_elem_list, xml_fun_inter_list, output_xml):
    """
    Check if each string in described_attribute_list are corresponding to an actual object and
    attribute, create new [Attribute, (Object, value)] objects list for object's type : Function
    and Functional Element.
    Send lists to add_object_attribute() to write them within xml and then returns update_list
    from it.

        Parameters:
            described_attribute_list ([str]) : Lists of string from jarvis cell
            xml_attribute_list ([Attribute]) : Attribute's list from xml
            xml_function_list ([Function]) : Function list from xml parsing
            xml_fun_elem_list ([Fun Elem]) : Functional Element list from xml parsing
            xml_fun_inter_list ([FunctionalInterface]) : FunctionalInterface list from xml parsing
            output_xml (GenerateXML object) : XML's file object

        Returns:
            update ([0/1]) : 1 if update, else 0
    """
    new_described_attribute_list = []
    # Create objects names/aliases list
    xml_attribute_name_list = get_objects_names(xml_attribute_list)
    xml_function_name_list = get_objects_names(xml_function_list)
    xml_fun_elem_name_list = get_objects_names(xml_fun_elem_list)
    xml_fun_inter_name_list = get_objects_names(xml_fun_inter_list)
    whole_list = xml_function_name_list + xml_fun_elem_name_list + xml_fun_inter_name_list

    # Loop to filter attributes and create a new list
    for elem in described_attribute_list:
        is_elem_found = True
        if not any(item == elem[1] for item in whole_list) and \
                not any(item == elem[0] for item in xml_attribute_name_list):
            is_elem_found = False
            print(f"{elem[1]} and {elem[0]} do not exist")
        elif not any(item == elem[1] for item in whole_list) or \
                not any(item == elem[0] for item in xml_attribute_name_list):
            is_elem_found = False
            if any(item == elem[1] for item in whole_list) and \
                    not any(item == elem[0] for item in xml_attribute_name_list):
                print(f"{elem[0]} does not exist")
            elif any(item == elem[0] for item in xml_attribute_name_list) and not \
                    any(item == elem[1] for item in whole_list):
                print(f"{elem[1]} does not exist")

        if is_elem_found:
            current_attrib = None
            for attribute in xml_attribute_list:
                if elem[0] == attribute.name or elem[0] == attribute.alias:
                    current_attrib = attribute
            # Loop to filter attribute and create a new list
            result_function = any(item == elem[1] for item in xml_function_name_list)
            result_fun_elem = any(item == elem[1]for item in xml_fun_elem_name_list)
            result_fun_inter = any(item == elem[1]for item in xml_fun_inter_name_list)

            if result_function and current_attrib:
                for function in xml_function_list:
                    if elem[1] == function.name or elem[1] == function.alias:
                        if (function.id, elem[2]) not in current_attrib.described_item_list:
                            new_described_attribute_list.append(
                                [current_attrib, (function, str(elem[2]))])

            if result_fun_elem and current_attrib:
                for fun_elem in xml_fun_elem_list:
                    if elem[1] == fun_elem.name or elem[1] == fun_elem.alias:
                        if (fun_elem.id, elem[2]) not in current_attrib.described_item_list:
                            new_described_attribute_list.append(
                                [current_attrib, (fun_elem, str(elem[2]))])

            if result_fun_inter and current_attrib:
                for fun_inter in xml_fun_inter_list:
                    if elem[1] == fun_inter.name or elem[1] == fun_inter.alias:
                        if (fun_inter.id, elem[2]) not in current_attrib.described_item_list:
                            new_described_attribute_list.append(
                                [current_attrib, (fun_inter, str(elem[2]))])

    update = add_object_attribute(new_described_attribute_list, output_xml)

    return update


def add_object_attribute(new_obj_attribute_list, output_xml):
    """
    Check if input list is not empty, write in xml for each element and return update list if some
    updates has been made

        Parameters:
            new_obj_attribute_list ([Attribute, (Object, value)]) : New described attributes
            output_xml (GenerateXML object) : XML's file object

        Returns:
            1 if update, else 0
    """
    if not new_obj_attribute_list:
        return 0

    output_xml.write_described_attribute_item(new_obj_attribute_list)
    # Warn the user once added within xml
    for described_attribute in new_obj_attribute_list:
        described_attribute[0].add_described_item(described_attribute[1])
        print(f"Attribute {described_attribute[0].name} for {described_attribute[1][0].name} "
              f"with value {described_attribute[1][1]}")
    return 1


def check_set_extends(extends_str_list, xml_type_list, output_xml):
    """
    Check if each type_to_extend string in extends_str_list are corresponding to actual objects
    name/alias, create lists for all <type> objects that needs to be added.

        Parameters:
            extends_str_list ([str]) : Lists of string from jarvis cell
            xml_type_list ([Type]) : xml list of type
            output_xml : xml's file object

        Returns:
            update ([0/1]) : 1 if update, else 0
    """
    new_type_list = []
    # Capitalyze the reference type for datamodel matching
    for elem in extends_str_list:
        if any(t == elem[0] for t in get_objects_names(xml_type_list)):
            # print(f"{elem[0]} already exists")
            continue
        type_to_extend = check_get_type_to_extend(elem[1], xml_type_list)
        if not type_to_extend:
            print(f"Enable to find referenced type '{elem[1]}'")
            continue
        new_type = datamodel.Type()
        new_type.set_name(elem[0])
        # Generate and set unique identifier of length 10 integers
        identifier = uuid.uuid4()
        new_type.set_id(str(identifier.int)[:10])
        if isinstance(type_to_extend, datamodel.Type):
            new_type.set_base(type_to_extend)
        else:
            new_type.set_base(type_to_extend.get_enum(elem[1].capitalize()))
        new_type_list.append(new_type)
        xml_type_list.add(new_type)

    if not new_type_list:
        return 0

    output_xml.write_type_element(new_type_list)
    for obj_type in new_type_list:
        if isinstance(obj_type.base, datamodel.Type):
            base_type = obj_type.base.name
        else:
            base_type = obj_type.base
        print(f"{obj_type.name} is a type extending {str(base_type)}")
    return 1


def check_get_type_to_extend(type_str, xml_type_list):
    """Checks if type_str is within BaseType or xml_type_list, then return Basetype or
    type object"""
    check = None
    if any(a == type_str.capitalize() for a in [str(i) for i in datamodel.BaseType]):
        check = datamodel.BaseType
        return check

    if any(a == type_str for a in get_objects_names(xml_type_list)):
        check = check_get_object(type_str, **{'xml_type_list': xml_type_list})
        return check

    return check



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Module for xml parsing"""
# Libraries
from lxml import etree

# Modules
import datamodel


class XmlParser3SE:
    """Class for xml parsing"""
    def __init__(self):
        self.xml_dict = {'xml_function_list': set(),
                         'xml_consumer_function_list': [],
                         'xml_producer_function_list': [],
                         'xml_data_list': set(),
                         'xml_state_list': set(),
                         'xml_transition_list': set(),
                         'xml_fun_elem_list': set(),
                         'xml_view_list': set(),
                         'xml_attribute_list': set(),
                         'xml_fun_inter_list': set(),
                         'xml_phy_elem_list': set(),
                         'xml_phy_inter_list': set(),
                         'xml_type_list': set()}
        self.root = None

    def parse_xml(self, input_filename):
        """Parses the whole xml then returns lists of objects/relationship"""
        # To speed up parsing (see lxml doc) : TBC if can be extended to xml_writer
        parser = etree.XMLParser(collect_ids=False)
        # Parse the XML file
        tree = etree.parse(input_filename, parser)
        # Get the XML tree
        self.root = tree.getroot()
        # Check xml root tag
        if not check_xml(self.root):
            user_msg = f"Xml's file structure has changed, please delete {input_filename} " \
                       f"and re-execute your whole notebook"
            return user_msg

        self.xml_dict = {'xml_type_list': get_type_list(self.root),
                         'xml_function_list': get_functions(self.root),
                         'xml_state_list': get_state(self.root),
                         'xml_transition_list': get_transition(self.root),
                         'xml_fun_elem_list': get_functional_element(self.root),
                         'xml_view_list': get_views(self.root),
                         'xml_attribute_list': get_attributes(self.root),
                         'xml_fun_inter_list': get_functional_interface(self.root),
                         'xml_phy_elem_list': get_physical_element(self.root),
                         'xml_phy_inter_list': get_physical_interface(self.root),
                         }
        # Create data(and set predecessors), consumers, producers lists
        self.xml_dict['xml_data_list'], self.xml_dict['xml_producer_function_list'], self.xml_dict[
            'xml_consumer_function_list'] = get_data(self.root, self.xml_dict['xml_function_list'])

        return self.xml_dict


def check_xml(root):
    """Check xml file root, since jarvis4se version 1.3 it's <systemAnalysis>"""
    if root.tag == "systemAnalysis":
        return True
    else:
        return False


def get_functions(root):
    """Get Function objects"""
    function_list = set()
    parent_list = {}
    xml_function_list = root.iter('function')
    for xml_function in xml_function_list:
        # Instantiate functions and add them to a list
        function = datamodel.Function(p_id=xml_function.get('id'), p_name=xml_function.get('name'),
                                      p_alias=xml_function.get('alias'),
                                      p_type=xml_function.get('type'),
                                      p_derived=xml_function.get('derived'))
        function.set_operand()

        function_list.add(function)
        # Looking for functions with "functionalPart" i.e childs and create a list
        xml_function_part_list = xml_function.iter('functionPart')
        for xml_function_part in xml_function_part_list:
            parent_list[xml_function_part.get('id')] = function.id

    # Loop to set parent() and add_child() to functions
    for child_id in parent_list:
        for child_function in function_list:
            if child_function.id == child_id:
                for parent_function in function_list:
                    if parent_function.id == parent_list[child_id]:
                        child_function.set_parent(parent_function)
                        parent_function.add_child(child_function)
                        break
                break

    for elem in function_list:
        for derived in function_list:
            if elem.derived == derived.id:
                elem.derived = derived
                break

    return function_list


def get_data(root, function_list):
    """Get Data objects"""
    data_list = set()
    consumer_function_list = []
    producer_function_list = []

    xml_data_list = root.iter('data')
    for xml_data in xml_data_list:
        # Instantiate data and add it to a list
        data = datamodel.Data(p_id=xml_data.get('id'),
                              p_name=xml_data.get('name'),
                              p_type=xml_data.get('type'))
        data_list.add(data)
        # looking for all elements with tag "consumer" and create a list [flow, consumer_function]
        xml_consumer_list = xml_data.iter('consumer')
        for xml_consumer in xml_consumer_list:
            for function in function_list:
                if xml_consumer.get('id') == function.id:
                    consumer_function_list.append([xml_data.get('name'), function])
                    if xml_consumer.get('role') != 'none':
                        function.set_input_role(xml_data.get('name'))
                    # Avoid to reset the input role once already set
                    elif function.input_role is None:
                        function.set_input_role(None)

        # looking for all elements with tag "producer" and create a list [flow, producer_function]
        xml_producer_list = xml_data.iter('producer')
        for xml_producer in xml_producer_list:
            for function in function_list:
                if xml_producer.get('id') == function.id:
                    producer_function_list.append([xml_data.get('name'), function])

    # Loop on the data_list once created to find the predecessor and add it to list
    for d in data_list:
        xml_data_list = root.iter('data')
        for xml_data in xml_data_list:
            if xml_data.get('id') == d.id:
                # looking for all elements with tag "predecessor" and create a
                # list [flow, producer_function]
                xml_predecessor_list = xml_data.iter('predecessor')
                for xml_predecessor in xml_predecessor_list:
                    for dodo in data_list:
                        if xml_predecessor.get('id') == dodo.id:
                            d.add_predecessor(dodo)

    return data_list, producer_function_list, consumer_function_list


def get_state(root):
    """Get State objects"""
    state_list = set()
    state_parent_dict = {}
    xml_state_list = root.iter('state')
    for xml_state in xml_state_list:
        # Instantiate states and add them to a list
        state = datamodel.State(p_id=xml_state.get('id'),
                                p_name=xml_state.get('name'),
                                p_alias=xml_state.get('alias'),
                                p_type=xml_state.get('type'))
        state_list.add(state)
        # Looking for states with "statePart" i.e child and create a list
        xml_state_part_list = xml_state.iter('statePart')
        for xml_state_part in xml_state_part_list:
            state_parent_dict[xml_state_part.get('id')] = state.id

        # Looking for allocated functions and add them to the state
        xml_allocated_function_list = xml_state.iter('allocatedFunction')
        for xml_allo_fun in xml_allocated_function_list:
            state.add_allocated_function(xml_allo_fun.get("id"))

    # Loop to set parent() and add_child() to states
    for child_id in state_parent_dict:
        for child_state in state_list:
            if child_state.id == child_id:
                for parent_state in state_list:
                    if parent_state.id == state_parent_dict[child_id]:
                        child_state.set_parent(parent_state)
                        parent_state.add_child(child_state)
                        break
                break

    return state_list


def get_transition(root):
    """Get Transition objects"""
    transition_list = set()
    xml_transition_list = root.iter('transition')
    for xml_transition in xml_transition_list:
        # Instantiate transitions and add them to a list
        transition = datamodel.Transition(p_id=xml_transition.get('id'),
                                          p_name=xml_transition.get('name'),
                                          p_alias=xml_transition.get('alias'),
                                          p_type=xml_transition.get('type'),
                                          p_source=xml_transition.get('source'),
                                          p_destination=xml_transition.get('destination'))
        # Looking for conditions and add them to the transition
        xml_transition_condition_list = xml_transition.iter('condition')
        for xml_condition in xml_transition_condition_list:
            transition.add_condition(xml_condition.get("text"))

        transition_list.add(transition)

    return transition_list


def get_functional_element(root):
    """Get Functional Element objects"""
    functional_element_list = set()
    fun_elem_parent_dict = {}
    xml_functional_element_list = root.iter('functionalElement')
    for xml_func_elem in xml_functional_element_list:
        # Instantiate functional element and add them to a list
        fun_elem = datamodel.FunctionalElement(p_id=xml_func_elem.get('id'),
                                               p_name=xml_func_elem.get('name'),
                                               p_alias=xml_func_elem.get('alias'),
                                               p_type=xml_func_elem.get('type'),
                                               p_derived=xml_func_elem.get('derived'))
        functional_element_list.add(fun_elem)
        # Looking for "functionalElementPart" i.e child and create a list
        xml_functional_part_list = xml_func_elem.iter('functionalElementPart')
        for xml_state_part in xml_functional_part_list:
            fun_elem_parent_dict[xml_state_part.get('id')] = fun_elem.id

        # Looking for allocated states and add them to the functional element
        xml_allocated_state_list = xml_func_elem.iter('allocatedState')
        for xml_allo_state in xml_allocated_state_list:
            fun_elem.add_allocated_state(xml_allo_state.get("id"))

        # Looking for allocated functions and add them to the functional element
        xml_allocated_function_list = xml_func_elem.iter('allocatedFunction')
        for xml_allo_fun in xml_allocated_function_list:
            fun_elem.add_allocated_function(xml_allo_fun.get("id"))

        # Looking for exposed interface and add them to the functional element
        xml_exposed_interface_list = xml_func_elem.iter('exposedInterface')
        for xml_exp_inter in xml_exposed_interface_list:
            fun_elem.add_exposed_interface(xml_exp_inter.get("id"))

    # Loop to set parent() and add_child() to fun elem
    for child_id in fun_elem_parent_dict:
        for child_state in functional_element_list:
            if child_state.id == child_id:
                for parent_state in functional_element_list:
                    if parent_state.id == fun_elem_parent_dict[child_id]:
                        child_state.set_parent(parent_state)
                        parent_state.add_child(child_state)
                        break
                break

    for elem in functional_element_list:
        for derived in functional_element_list:
            if elem.derived == derived.id:
                elem.derived = derived
                break

    return functional_element_list


def get_views(root):
    """Get View objects"""
    view_list = set()
    xml_view_list = root.iter('view')
    for xml_view in xml_view_list:
        # Instantiate view and add them to a list
        view = datamodel.View(p_id=xml_view.get('id'),
                              p_name=xml_view.get('name'),
                              p_type=xml_view.get('type'))
        # Looking for allocated items and add them to the view
        xml_allocated_item_list = xml_view.iter('allocatedItem')
        for xml_allo_item in xml_allocated_item_list:
            view.add_allocated_item(xml_allo_item.get("id"))

        view_list.add(view)

    return view_list


def get_attributes(root):
    """Get Attribute objects"""
    attribute_list = set()
    xml_attribute_list = root.iter('attribute')
    for xml_attribute in xml_attribute_list:
        # Instantiate Attribute and add them to a list
        attribute = datamodel.Attribute(p_id=xml_attribute.get('id'),
                                        p_name=xml_attribute.get('name'),
                                        p_alias=xml_attribute.get('alias'),
                                        p_type=xml_attribute.get('type'))
        # Looking for described items and add them to the attribute
        xml_described_item_list = xml_attribute.iter('describedItem')
        for xml_described_item in xml_described_item_list:
            attribute.add_described_item((xml_described_item.get("id"),
                                          xml_described_item.get("value")))

        attribute_list.add(attribute)

    return attribute_list


def get_functional_interface(root):
    """Get Functional Interface objects"""
    functional_interface_list = set()
    xml_fun_inter_list = root.iter('functionalInterface')
    for xml_fun_inter in xml_fun_inter_list:
        # Instantiate fun_inter and add them to a list
        fun_inter = datamodel.FunctionalInterface(p_id=xml_fun_inter.get('id'),
                                                  p_name=xml_fun_inter.get('name'),
                                                  p_alias=xml_fun_inter.get('alias'),
                                                  p_type=xml_fun_inter.get('type'),
                                                  p_derived=xml_fun_inter.get('derived'))
        # Looking for allocated data and add them to the fun inter
        xml_allocated_data_list = xml_fun_inter.iter('allocatedData')
        for xml_allo_data in xml_allocated_data_list:
            fun_inter.add_allocated_data(xml_allo_data.get("id"))

        functional_interface_list.add(fun_inter)

    for elem in functional_interface_list:
        for derived in functional_interface_list:
            if elem.derived == derived.id:
                elem.derived = derived
                break

    return functional_interface_list


def get_physical_element(root):
    """Get Physical Element objects"""
    physical_element_list = set()
    phy_elem_parent_dict = {}
    xml_physical_element_list = root.iter('physicalElement')
    for xml_phy_elem in xml_physical_element_list:
        # Instantiate functional element and add them to a list
        phy_elem = datamodel.PhysicalElement(p_id=xml_phy_elem.get('id'),
                                             p_name=xml_phy_elem.get('name'),
                                             p_alias=xml_phy_elem.get('alias'),
                                             p_type=xml_phy_elem.get('type'),
                                             p_derived=xml_phy_elem.get('derived'))
        physical_element_list.add(phy_elem)
        # Looking for "physicalPart" i.e child and create a list
        xml_functional_part_list = xml_phy_elem.iter('physicalElementPart')
        for xml_state_part in xml_functional_part_list:
            phy_elem_parent_dict[xml_state_part.get('id')] = phy_elem.id

        # Looking for allocated functions and add them to the functional element
        xml_allocated_fun_elem_list = xml_phy_elem.iter('allocatedFunctionalElement')
        for xml_allo_fun_elem in xml_allocated_fun_elem_list:
            phy_elem.add_allocated_fun_elem(xml_allo_fun_elem.get("id"))

        # Looking for exposed interface and add them to the functional element
        xml_exposed_interface_list = xml_phy_elem.iter('exposedInterface')
        for xml_exp_inter in xml_exposed_interface_list:
            phy_elem.add_exposed_interface(xml_exp_inter.get("id"))

    # Loop to set parent() and add_child() to fun elem
    for child_id in phy_elem_parent_dict:
        for child_state in physical_element_list:
            if child_state.id == child_id:
                for parent_state in physical_element_list:
                    if parent_state.id == phy_elem_parent_dict[child_id]:
                        child_state.set_parent(parent_state)
                        parent_state.add_child(child_state)
                        break
                break

    for elem in physical_element_list:
        for derived in physical_element_list:
            if elem.derived == derived.id:
                elem.derived = derived
                break

    return physical_element_list


def get_physical_interface(root):
    """Get Physical Interface objects"""
    physical_interface_list = set()
    xml_phy_inter_list = root.iter('physicalInterface')
    for xml_phy_inter in xml_phy_inter_list:
        # Instantiate phy_inter and add them to a list
        phy_inter = datamodel.PhysicalInterface(p_id=xml_phy_inter.get('id'),
                                                p_name=xml_phy_inter.get('name'),
                                                p_alias=xml_phy_inter.get('alias'),
                                                p_type=xml_phy_inter.get('type'),
                                                p_derived=xml_phy_inter.get('derived'))
        # Looking for allocated fun_inter and add them to the phy inter
        xml_allocated_inter_list = xml_phy_inter.iter('allocatedFunctionalInterface')
        for xml_allo_inter in xml_allocated_inter_list:
            phy_inter.add_allocated_fun_inter(xml_allo_inter.get("id"))

        physical_interface_list.add(phy_inter)

    for inter in physical_interface_list:
        for derived in physical_interface_list:
            if inter.derived == derived.id:
                inter.derived = derived
                break

    return physical_interface_list


def get_type_list(root):
    """Get Type objects"""
    type_list = set()
    xml_type_list = root.iter('type')
    for xml_type in xml_type_list:
        # Instantiate Type and add them to a list
        type_obj = datamodel.Type(p_id=xml_type.get('id'),
                                  p_name=xml_type.get('name'),
                                  p_alias=xml_type.get('alias'),
                                  p_base=xml_type.get('base'))
        type_list.add(type_obj)

    for obj_type in type_list:
        for base in type_list:
            if obj_type.base == base.name:
                obj_type.base = base
                break

    return type_list



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Libraries
import re

import datamodel
from .question_answer import get_objects_names
from . import shared_orchestrator


def create_phy_elem_obj(phy_elem_str, specific_obj_type, **kwargs):
    """
    Check if string phy_elem_str is not already corresponding to an actual object's name/alias,
    create new PhysicalElement() object, instantiate it, add it to xml_phy_elem_list.

        Parameters:
            phy_elem_str (str) : Lists of string from jarvis cell
            specific_obj_type ([Function]) : specific type or None
            kwargs (dict) : whole xml lists + xml's file object

        Returns:
            update (0/fun_elem) : fun_elem if update, else 0
    """
    if any(n == phy_elem_str for n in get_objects_names(kwargs['xml_phy_elem_list'])):
        return 0
    if not specific_obj_type:
        phy_elem = datamodel.PhysicalElement(p_name=phy_elem_str)
    else:
        phy_elem = datamodel.PhysicalElement(p_name=phy_elem_str, p_type=specific_obj_type.name)

    alias_str = re.search(r"(.*)\s[-]\s", phy_elem_str, re.MULTILINE)
    if alias_str:
        phy_elem.set_alias(alias_str.group(1))
    phy_elem.set_id(shared_orchestrator.get_unique_id())

    kwargs['xml_phy_elem_list'].add(phy_elem)
    print(f"{phy_elem.name} is a {phy_elem.type}")
    if phy_elem:
        return phy_elem
    return 0


def create_phy_inter_obj(phy_inter_str, specific_obj_type, **kwargs):
    """
    Check if string phy_inter_str is not already corresponding to an actual object's name/alias,
    create new PhysicalElement() object, instantiate it, add it to xml_phy_inter_list.

        Parameters:
            phy_inter_str (str) : Lists of string from jarvis cell
            specific_obj_type ([Function]) : specific type or None
            kwargs (dict) : whole xml lists + xml's file object

        Returns:
            update (0/fun_elem) : fun_elem if update, else 0
    """
    if any(n == phy_inter_str for n in get_objects_names(kwargs['xml_phy_inter_list'])):
        return 0
    if not specific_obj_type:
        phy_inter = datamodel.PhysicalInterface(p_name=phy_inter_str)
    else:
        phy_inter = datamodel.PhysicalInterface(p_name=phy_inter_str, p_type=specific_obj_type.name)

    alias_str = re.search(r"(.*)\s[-]\s", phy_inter_str, re.MULTILINE)
    if alias_str:
        phy_inter.set_alias(alias_str.group(1))
    phy_inter.set_id(shared_orchestrator.get_unique_id())

    kwargs['xml_phy_inter_list'].add(phy_inter)
    print(f"{phy_inter.name} is a {phy_inter.type}")
    if phy_inter:
        return phy_inter
    return 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Module containing methods shared between objects/orchestrator"""
import re
import uuid

import datamodel
from .question_answer import (get_object_type, check_get_object, get_allocation_object,
                              check_not_family, get_objects_names, get_children)
from .functional_orchestrator import (create_function_obj, create_data_obj, create_fun_elem_obj,
                                      create_fun_inter_obj)
from .physical_orchestrator import (create_phy_elem_obj, create_phy_inter_obj)


def cut_string_list(string_tuple_list):
    """From set of input command strings e.g for composition with input list as
    {(Function_name, Function_name_A), (Function_name, [Function_name_B, Function_name_C])} : this
    methods returns {(Function_name, Function_name_A), (Function_name, Function_name_B),
    (Function_name, Function_name_C)}

        Parameters:
            string_tuple_list ([(str, str), ...]) : Lists of string tuple from jarvis cell
        Returns:
            output_list ([0/1]) : output list
    """

    output_list = []
    for parent, child in string_tuple_list:
        if "," in child:
            child_str = child.replace(" ", "")
            child_list_str = re.split(r',(?![^[]*\])', child_str)
            for elem in child_list_str:
                output_list.append((parent, elem))
        else:
            output_list.append((parent, child))

    return output_list


def check_add_child(parent_child_name_str_list, **kwargs):
    """
    Check if each string in parent_child_name_str_list are corresponding to an actual object,
    create new [parent, child] objects lists for object's type : State/Function/FunctionalElement.
    Send lists to add_child() to write them within xml and then returns update_list from it.

        Parameters:
            parent_child_name_str_list ([str]) : Lists of string from jarvis cell
            kwargs (dict) : 4 xml lists(see matched_composition() within command_parser.py)
            + xml's file object

        Returns:
            update ([0/1]) : 1 if update, else 0
    """

    parent_child_lists = [[] for _ in range(4)]
    available_objects = (datamodel.Function, datamodel.State,
                         datamodel.FunctionalElement,  datamodel.PhysicalElement)

    cleaned_parent_child_list_str = cut_string_list(parent_child_name_str_list)
    for elem in cleaned_parent_child_list_str:
        parent_object = check_get_object(elem[0], **kwargs)
        child_object = check_get_object(elem[1], **kwargs)
        if parent_object is None:
            if child_object is None:
                print(f"{elem[0]} and {elem[1]} are not Function/State/FunctionalElement"
                      f"/PhysicalElement or the object does not exist")
                continue

            print(f"{elem[0]} is not Function/State/FunctionalElement"
                  f"/PhysicalElement or the object does not exist")
            continue
        if child_object is None:
            print(f"{elem[1]} is not Function/State/FunctionalElement"
                  f"/PhysicalElement or the object does not exist")
            continue
        check_pair = None
        for idx, obj_type in enumerate(available_objects):
            if isinstance(parent_object, obj_type) and isinstance(child_object, obj_type):
                check_pair = idx
                break
        if isinstance(check_pair, int):
            if child_object.parent is None:
                parent_object.add_child(child_object)
                child_object.set_parent(parent_object)
                parent_child_lists[check_pair].append([parent_object, child_object])
        else:
            print(f"Please choose a valid pair of element(Function/State/FunctionalElement"
                  f"/PhysicalElement) for {parent_object.name} and {child_object.name}")

    update = add_child(parent_child_lists, kwargs['xml_fun_elem_list'], kwargs['output_xml'])

    return update


def add_child(parent_child_lists, xml_fun_elem_list, output_xml):
    """
    Check if input lists are not empty, write in xml for each list and return update list if some
    updates has been made

        Parameters:
            parent_child_lists ([Parent, Child]) : [[Function],[State],[FunctionalElement],
            [PhysicalElement]]
            xml_fun_elem_list ([FunctionalElement]) : functional element list from xml parsing
            output_xml (GenerateXML object) : XML's file object

        Returns:
            update_list ([0/1]) : Add 1 to list if any update, otherwise 0 is added
    """
    if any(parent_child_lists):
        for i in range(4):
            if parent_child_lists[i]:
                output_xml.write_object_child(parent_child_lists[i])
                for k in parent_child_lists[i]:
                    print(f"{k[0].name} is composed of {k[1].name}")
                    if i in (0, 1):
                        for fun_elem in xml_fun_elem_list:
                            if k[0].id in fun_elem.allocated_function_list:
                                recursive_allocation([fun_elem, k[1]], output_xml)
        return 1

    return 0


def check_add_allocated_item(item, xml_item_list, xml_view_list):
    """
    Checks if a view is already activated, if yes check if item isn't already
    allocated and returns corresponding [View, Object].
    Args:
        item (string): Object's name/alias from user's input
        xml_item_list ([Object]): List of xml's item (same type as item)
        xml_view_list ([View]) : View list from xml parsing

    Returns:
        [View, Object]
    """
    if not any(s.activated for s in xml_view_list):
        return None

    activated_view = None
    for view in xml_view_list:
        if view.activated:
            activated_view = view
            break
    if activated_view:
        for i in xml_item_list:
            if item == i.name:
                if i.id not in activated_view.allocated_item_list:
                    activated_view.add_allocated_item(i.id)
                    return [activated_view, i]
            # To avoid errors for i.alias when i is Data (no such attriute)
            try:
                if item == i.alias:
                    if i.id not in activated_view.allocated_item_list:
                        activated_view.add_allocated_item(i.id)
                        return [activated_view, i]
            except AttributeError:
                pass


def check_and_delete_object(delete_str_list, **kwargs):
    """
    Check if each string in delete_str_list are corresponding to an actual object, create new
    objects list for objects to delete.
    Send lists to delete_objects() to delete them within xml and then returns update from it.

        Parameters:
            delete_str_list ([str]) : List of string from jarvis cell
            kwargs (dict) : whole xml lists + xml's file object

        Returns:
            update ([0/1]) : 1 if update, else 0
    """
    to_be_deleted_obj_lists = [[] for _ in range(10)]
    # Check if the wanted to delete object exists and can be deleted
    for obj_str in delete_str_list:
        object_to_del = check_get_object(obj_str, **kwargs)
        if object_to_del is None:
            print(f"{obj_str} does not exist")
            continue
        check, list_idx = check_relationship_before_delete(object_to_del, **kwargs)
        if check:
            to_be_deleted_obj_lists[list_idx].append(object_to_del)
        else:
            print(f"{object_to_del.name} can not be deleted")

    update = delete_objects(to_be_deleted_obj_lists, kwargs['output_xml'])

    return update


def check_relationship_before_delete(object_to_del, **kwargs):
    """Switch to trigger differents methods depend object's type"""
    _, idx = get_specific_obj_type_and_idx(object_to_del)
    switch_check = {
        0: check_function,
        1: check_data,
        2: check_state,
        3: check_transition,
        4: check_fun_elem,
        5: check_chain,
        6: check_attribute,
        7: check_fun_inter,
        8: check_phy_elem,
        9: check_phy_inter,
    }
    check_obj = switch_check.get(idx, "Object can not be deleted")
    check = check_obj(object_to_del, **kwargs)
    return check, idx


def check_object_not_in_prod_cons(object_to_check, consumer_list, producer_list):
    """Check if object/data_name not in producer or consumer lists"""
    check = False
    if not any(object_to_check in o for o in consumer_list + producer_list):
        check = True
    return check


def check_object_no_parent_and_child(object_to_check):
    """Check if an object has not parent and not child"""
    check = False
    if not object_to_check.child_list and object_to_check.parent is None:
        check = True

    return check


def check_object_not_allocated(object_to_check, allocated_to_object_list):
    """Check if object in allocated_list of allocated objects"""
    check = False
    if not allocated_to_object_list:
        check = True
        return check

    converted_list = list(allocated_to_object_list)
    if isinstance(object_to_check, datamodel.Function):
        if isinstance(converted_list[0], (datamodel.State, datamodel.FunctionalElement)):
            if not any(object_to_check.id in obj.allocated_function_list for obj
                       in converted_list):
                check = True
        if isinstance(converted_list[0], datamodel.View):
            if not any(object_to_check.id in obj.allocated_item_list for obj
                       in converted_list):
                check = True
    if isinstance(object_to_check, datamodel.Data):
        if isinstance(converted_list[0], datamodel.View):
            if not any(object_to_check.id in obj.allocated_item_list for obj
                       in converted_list):
                check = True
        if isinstance(converted_list[0], datamodel.FunctionalInterface):
            if not any(object_to_check.id in obj.allocated_data_list for obj
                       in converted_list):
                check = True
    if isinstance(object_to_check, datamodel.State):
        if isinstance(converted_list[0], datamodel.FunctionalElement):
            if not any(object_to_check.id in obj.allocated_state_list for obj
                       in converted_list):
                check = True
    if isinstance(object_to_check, datamodel.FunctionalElement):
        if isinstance(converted_list[0], datamodel.PhysicalElement):
            if not any(object_to_check.id in obj.allocated_fun_elem_list for obj
                       in converted_list):
                check = True
    if isinstance(object_to_check, datamodel.FunctionalInterface):
        if isinstance(converted_list[0], datamodel.FunctionalElement):
            if not any(object_to_check.id in obj.exposed_interface_list for obj
                       in converted_list):
                check = True
        if isinstance(converted_list[0], datamodel.PhysicalInterface):
            if not any(object_to_check.id in obj.allocated_fun_inter_list for obj
                       in converted_list):
                check = True
    if isinstance(object_to_check, datamodel.PhysicalInterface):
        if isinstance(converted_list[0], datamodel.PhysicalElement):
            if not any(object_to_check.id in obj.exposed_interface_list for obj
                       in converted_list):
                check = True
    return check


def check_object_no_attribute(object_to_check, attribute_list):
    """Check that object has not attribute set"""
    check = False
    described_item_list = [obj.described_item_list for obj in attribute_list]
    if not any(object_to_check.id in obj for obj in described_item_list):
        check = True
    return check


def check_function(object_to_del, **kwargs):
    """Checks for Function's object"""
    check = False
    check_list = [False]*6
    check_list[0] = check_object_no_parent_and_child(object_to_del)
    if not check_list[0]:
        print(f"{object_to_del.name} has composition relationship(s)")

    check_list[1] = check_object_not_allocated(object_to_del, kwargs['xml_state_list'])
    check_list[2] = check_object_not_allocated(object_to_del, kwargs['xml_fun_elem_list'])
    if not check_list[1] or not check_list[2]:
        print(f"{object_to_del.name} has allocation relationship(s)")

    check_list[3] = check_object_not_in_prod_cons(object_to_del,
                                                  kwargs['xml_consumer_function_list'],
                                                  kwargs['xml_producer_function_list'])
    if not check_list[3]:
        print(f"{object_to_del.name} has production/consumption relationship(s)")

    check_list[4] = check_object_no_attribute(object_to_del, kwargs['xml_attribute_list'])
    if not check_list[4]:
        print(f"{object_to_del.name} has attribute(s) set")

    check_list[5] = check_object_not_allocated(object_to_del, kwargs['xml_view_list'])
    if not check_list[5]:
        print(f"{object_to_del.name} has chain relationship(s)")

    if all(check_list):
        check = True
        kwargs['xml_function_list'].remove(object_to_del)
    return check


def check_data(object_to_del, **kwargs):
    """Checks for Data's object"""
    check = False
    check_list = [False]*3
    check_list[0] = check_object_not_in_prod_cons(object_to_del.name,
                                                  kwargs['xml_consumer_function_list'],
                                                  kwargs['xml_producer_function_list'])
    if not check_list[0]:
        print(f"{object_to_del.name} has production/consumption relationship(s)")

    check_list[1] = check_object_not_allocated(object_to_del, kwargs['xml_view_list'])
    if not check_list[1]:
        print(f"{object_to_del.name} has chain relationship(s)")

    check_list[2] = check_object_not_allocated(object_to_del, kwargs['xml_fun_inter_list'])
    if not check_list[2]:
        print(f"{object_to_del.name} has allocation relationship(s)")

    if all(check_list):
        check = True
        kwargs['xml_data_list'].remove(object_to_del)
    return check


def check_state(object_to_del, **kwargs):
    """Checks for State's object"""
    check = False
    check_list = [False] * 4

    check_list[0] = check_object_no_parent_and_child(object_to_del)
    if not check_list[0]:
        print(f"{object_to_del.name} has composition relationship(s)")

    check_list[1] = not object_to_del.allocated_function_list
    check_list[2] = check_object_not_allocated(object_to_del, kwargs['xml_fun_elem_list'])
    if not check_list[1] or not check_list[2]:
        print(f"{object_to_del.name} has allocation relationship(s)")

    check_list[3] = not any(object_to_del.id in (trans.source, trans.destination)
                            for trans in kwargs['xml_transition_list'])
    if not check_list[3]:
        print(f"{object_to_del.name} has transition relationship(s)")
    if all(check_list):
        check = True
        kwargs['xml_state_list'].remove(object_to_del)
    return check


def check_transition(object_to_del, **kwargs):
    """Checks for State's object"""
    check = False
    check_list = [False] * 1

    check_list[0] = object_to_del.source is None and object_to_del.destination is None
    if not check_list[0]:
        print(f"{object_to_del.name} has source/destination relationship(s)")

    if all(check_list):
        check = True
        kwargs['xml_transition_list'].remove(object_to_del)
    return check


def check_fun_elem(object_to_del, **kwargs):
    """Checks for Functional Element's object"""
    check = False
    check_list = [False] * 4

    check_list[0] = check_object_no_parent_and_child(object_to_del)
    if not check_list[0]:
        print(f"{object_to_del.name} has composition relationship(s)")

    check_list[1] = (not object_to_del.allocated_function_list and
                     not object_to_del.allocated_state_list)
    check_list[2] = check_object_not_allocated(object_to_del, kwargs['xml_phy_elem_list'])
    if not check_list[1] or not check_list[2]:
        print(f"{object_to_del.name} has allocation relationship(s)")

    check_list[3] = not object_to_del.exposed_interface_list
    if not check_list[3]:
        print(f"{object_to_del.name} has interface relationship(s)")

    if all(check_list):
        check = True
        kwargs['xml_fun_elem_list'].remove(object_to_del)

    return check


def check_chain(object_to_del, **kwargs):
    """Checks for View's object"""
    check = False
    check_list = [False] * 1

    check_list[0] = not object_to_del.allocated_item_list
    if not check_list[0]:
        print(f"{object_to_del.name} has allocation relationship(s)")

    if all(check_list):
        check = True
        kwargs['xml_view_list'].remove(object_to_del)

    return check


def check_attribute(object_to_del, **kwargs):
    """Checks for Attribute's object"""
    check = False
    check_list = [False] * 1

    check_list[0] = not object_to_del.described_item_list
    if not check_list[0]:
        print(f"{object_to_del.name} has attribute relationship(s)")

    if all(check_list):
        check = True
        kwargs['xml_attribute_list'].remove(object_to_del)

    return check


def check_fun_inter(object_to_del, **kwargs):
    """Checks for Functional Interface's object"""
    check = False
    check_list = [False] * 3

    check_list[0] = not object_to_del.allocated_data_list
    check_list[1] = check_object_not_allocated(object_to_del, kwargs['xml_fun_elem_list'])
    check_list[2] = check_object_not_allocated(object_to_del, kwargs['xml_phy_inter_list'])
    if not check_list[0] or not check_list[1] or not check_list[2]:
        print(f"{object_to_del.name} has allocation relationship(s)")

    if all(check_list):
        check = True
        kwargs['xml_fun_inter_list'].remove(object_to_del)

    return check


def check_phy_elem(object_to_del, **kwargs):
    """Checks for Physical Element's object"""
    check = False
    check_list = [False] * 3

    check_list[0] = check_object_no_parent_and_child(object_to_del)
    if not check_list[0]:
        print(f"{object_to_del.name} has composition relationship(s)")

    check_list[1] = not object_to_del.allocated_fun_elem_list
    if not check_list[1]:
        print(f"{object_to_del.name} has allocation relationship(s)")

    check_list[2] = not object_to_del.exposed_interface_list
    if not check_list[2]:
        print(f"{object_to_del.name} has interface relationship(s)")

    if all(check_list):
        check = True
        kwargs['xml_phy_elem_list'].remove(object_to_del)

    return check


def check_phy_inter(object_to_del, **kwargs):
    """Checks for Physical Interface's object"""
    check = False
    check_list = [False] * 2

    check_list[0] = not object_to_del.allocated_fun_inter_list
    if not check_list[0]:
        print(f"{object_to_del.name} has allocation relationship(s)")

    check_list[1] = check_object_not_allocated(object_to_del, kwargs['xml_phy_elem_list'])
    if not check_list[1]:
        print(f"{object_to_del.name} has interface relationship(s)")

    if all(check_list):
        check = True
        kwargs['xml_phy_inter_list'].remove(object_to_del)

    return check


def delete_objects(object_lists, output_xml):
    """
    Check if input lists are not empty, write in xml for each list and return update list if some
    updates has been made

        Parameters:
            object_lists : see order in get_specific_obj_type_and_idx()
            output_xml (GenerateXML object) : XML's file object

        Returns:
            1 if update, else 0
    """
    if any(object_lists):
        for i in range(10):
            if object_lists[i]:
                output_xml.delete_object(object_lists[i])
                for object_type in object_lists[i]:
                    print(f"{object_type.name} deleted")
        return 1
    return 0


def check_set_object_type(type_str_list, **kwargs):
    """
    Check if each string in type_str_list are corresponding to an actual object's name/alias, create
    [objects] ordered lists for:
    [[Function],[Data],[State],[Transition],[FunctionalElement],[Attribute],
    [FuncitonalInterface],[PhysicalElement],[PhysicalInterface],[View]]
    Send lists to set_object_type() to write them within xml and then returns update from it.

        Parameters:
            type_str_list ([str]) : Lists of string from jarvis cell
            kwargs (dict) : whole xml lists + xml's file object

        Returns:
            update ([0/1]) : 1 if update, else 0
    """
    object_type_lists = [[] for _ in range(10)]
    # Check if the wanted object exists and the type can be set
    for object_str, type_name in type_str_list:
        object_to_set = check_get_object(object_str, **kwargs)
        if object_to_set is None:
            print(f"{object_str} does not exist")
            continue
        if object_to_set.type == type_name:
            continue
        check, list_idx = check_new_type(object_to_set, type_name, kwargs['xml_type_list'])
        if check:
            object_type_lists[list_idx].append(object_to_set)

    update = set_object_type(object_type_lists, kwargs['output_xml'])

    return update


def check_new_type(object_to_set, type_name, xml_type_list):
    """Check if type in specity object's type list and if changed"""
    check = False
    specific_obj_type_list, list_idx = get_specific_obj_type_and_idx(object_to_set)
    if list_idx in (0, 1, 2, 3, 4, 5, 7, 8, 9):
        if any(t == type_name.upper() for t in specific_obj_type_list):
            if type_name.capitalize() != str(object_to_set.type):
                check = True
                object_to_set.set_type(type_name.capitalize())
        elif any(t == type_name for t in get_objects_names(xml_type_list)):
            obj_type = check_get_object(type_name, **{'xml_type_list': xml_type_list})
            check = check_type_recursively(obj_type, specific_obj_type_list)
            if not check:
                print(f"{obj_type.name} is not base type: "
                      f"{specific_obj_type_list[0].capitalize()}")
            else:
                object_to_set.set_type(obj_type.name)
        else:
            print(
                f"The type {type_name} does not exist, available types are "
                f": {', '.join(specific_obj_type_list)}.")

    elif list_idx == 6:
        check = True
        object_to_set.set_type(type_name)

    return check, list_idx


def check_type_recursively(obj_type, specific_obj_type_list):
    """Checks type.base recursively if it within specific_obj_type_list"""
    check = False
    if isinstance(obj_type.base, str) and obj_type.base.upper() in specific_obj_type_list:
        check = True
        return check
    elif isinstance(obj_type.base, datamodel.Type):
        return check_type_recursively(obj_type.base, specific_obj_type_list)
    return check


def get_specific_obj_type_and_idx(object_to_set):
    """Get __str__ list from FunctionType, DataType, StateType, TransitionType,
    FunctionalElementType, ViewType and index for output_list (depends on the type)"""
    specific_obj_type_list = []
    list_idx = None
    if isinstance(object_to_set, datamodel.Function):
        specific_obj_type_list = [str(i).upper() for i in datamodel.FunctionType]
        list_idx = 0
    elif isinstance(object_to_set, datamodel.Data):
        specific_obj_type_list = [str(i).upper() for i in datamodel.DataType]
        list_idx = 1
    elif isinstance(object_to_set, datamodel.State):
        specific_obj_type_list = [str(i).upper() for i in datamodel.StateType]
        list_idx = 2
    elif isinstance(object_to_set, datamodel.Transition):  # TBC
        specific_obj_type_list = [str(i).upper() for i in datamodel.TransitionType]
        list_idx = 3
    elif isinstance(object_to_set, datamodel.FunctionalElement):
        specific_obj_type_list = [str(i).upper() for i in datamodel.FunctionalElementType]
        list_idx = 4
    elif isinstance(object_to_set, datamodel.View):
        specific_obj_type_list = [str(i).upper() for i in datamodel.ViewType]
        list_idx = 5
    elif isinstance(object_to_set, datamodel.Attribute):
        list_idx = 6
    elif isinstance(object_to_set, datamodel.FunctionalInterface):
        specific_obj_type_list = [str(datamodel.BaseType.FUNCTIONAL_INTERFACE).upper()]
        list_idx = 7
    elif isinstance(object_to_set, datamodel.PhysicalElement):
        specific_obj_type_list = [str(datamodel.BaseType.PHYSICAL_ELEMENT).upper()]
        list_idx = 8
    elif isinstance(object_to_set, datamodel.PhysicalInterface):
        specific_obj_type_list = [str(datamodel.BaseType.PHYSICAL_INTERFACE).upper()]
        list_idx = 9

    return specific_obj_type_list, list_idx


def set_object_type(object_lists, output_xml):
    """
    Check if input lists are not empty, write in xml for each list and return update list if some
    updates has been made

        Parameters:
            object_lists : see order in get_specific_obj_type_and_idx()
            output_xml (GenerateXML object) : XML's file object

        Returns:
            1 if update, else 0
    """
    if any(object_lists):
        for i in range(10):
            if object_lists[i]:
                output_xml.write_object_type(object_lists[i])
                for object_type in object_lists[i]:
                    print(f"The type of {object_type.name} is {object_type.type}")
        return 1
    return 0


def check_set_object_alias(alias_str_list, **kwargs):
    """
    Check if each string in alias_str_list are corresponding to an actual object's name/alias,
    create [objects] ordered lists for:
    [[Function],[State],[Transition],[FunctionalElement],[Attribute],
    [FuncitonalInterface],[PhysicalElement],[PhysicalInterface]]
    Send lists to set_object_type() to write them within xml and then returns update from it.

        Parameters:
            alias_str_list ([str]) : Lists of string from jarvis cell
            kwargs (dict) : whole xml lists + xml's file object

        Returns:
            update ([0/1]) : 1 if update, else 0
    """
    object_lists = [[] for _ in range(9)]
    # Check if the wanted to object exists and the type can be set
    for object_to_set_alias, alias_str in alias_str_list:
        object_to_set = check_get_object(object_to_set_alias, **kwargs)
        if object_to_set is None:
            print(f"{object_to_set_alias} does not exist")
            continue

        idx = check_new_alias(object_to_set, alias_str)
        if isinstance(idx, int):
            object_lists[idx].append(object_to_set)

    update = set_object_alias(object_lists, kwargs['output_xml'])

    return update


def check_new_alias(object_to_set, alias_str):
    """Check that alias is new and object has en alias attribute, then returns corresponding
    object's type index"""
    list_idx = None
    if object_to_set.alias != alias_str:
        object_to_set.set_alias(alias_str)
        if isinstance(object_to_set, datamodel.Function):
            list_idx = 0
        elif isinstance(object_to_set, datamodel.State):
            list_idx = 1
        elif isinstance(object_to_set, datamodel.Transition):
            list_idx = 2
        elif isinstance(object_to_set, datamodel.FunctionalElement):
            list_idx = 3
        elif isinstance(object_to_set, datamodel.Attribute):
            list_idx = 4
        elif isinstance(object_to_set, datamodel.FunctionalInterface):
            list_idx = 5
        elif isinstance(object_to_set, datamodel.PhysicalElement):
            list_idx = 6
        elif isinstance(object_to_set, datamodel.PhysicalInterface):
            list_idx = 7
        elif isinstance(object_to_set, datamodel.Type):
            list_idx = 8
        else:
            print(f"{object_to_set.name} does not have alias attribute")

    return list_idx


def set_object_alias(object_lists, output_xml):
    """
    Check if input lists are not empty, write in xml for each list and return update if some
    updates has been made.
        Parameters:
            object_lists ([Object]) : object with new alias
            output_xml (GenerateXML object) : XML's file object

        Returns:
            1 if update, else 0
    """
    if any(object_lists):
        for i in range(9):
            if object_lists[i]:
                output_xml.write_object_alias(object_lists[i])
                for object_alias in object_lists[i]:
                    print(f"The alias for {object_alias.name} is {object_alias.alias}")

        return 1

    return 0


def check_add_allocation(allocation_str_list, **kwargs):
    """
    Check if each string in allocation_str_list are corresponding to an actual object's name/alias,
    create lists for:
    [[FunctionalElement, Function/State], [FunctionalInterface, Data],[State, Function],
    [PhysicalElement, FunctionalElement], [PhysicalInterface, FunctionalInterface]]
    Send lists to add_allocation() to write them within xml and then returns update from it.

        Parameters:
            allocation_str_list ([str]) : Lists of string from jarvis cell
            kwargs (dict) : whole xml lists + xml's file object

        Returns:
            update ([0/1]) : 1 if update, else 0
    """
    new_allocation = {
        0: [],  # [FunctionalElement, Function/State]
        1: [],  # [State, Function]
        2: [],  # [FunctionalInterface, Data]
        3: [],  # [PhysicalElement, FunctionalElement]
        4: [],  # [PhysicalInterface, FunctionalInterface]
        # 5: [],  [View, Object] in other modules or [Fun_elem_Parent, Function/State] in
        # check_parent_allocation() it's just a key with no recursivety
    }
    for elem in allocation_str_list:
        alloc_obj = check_get_object(elem[0],
                                     **{'xml_fun_elem_list': kwargs['xml_fun_elem_list'],
                                        'xml_state_list': kwargs['xml_state_list'],
                                        'xml_fun_inter_list': kwargs['xml_fun_inter_list'],
                                        'xml_phy_elem_list': kwargs['xml_phy_elem_list'],
                                        'xml_phy_inter_list': kwargs['xml_phy_inter_list'],
                                        })
        obj_to_alloc = check_get_object(elem[1],
                                        **{'xml_function_list': kwargs['xml_function_list'],
                                            'xml_state_list': kwargs['xml_state_list'],
                                           'xml_data_list': kwargs['xml_data_list'],
                                           'xml_fun_elem_list': kwargs['xml_fun_elem_list'],
                                           'xml_fun_inter_list': kwargs['xml_fun_inter_list'],
                                           })
        check_obj = check_allocation_objects_types(alloc_obj, obj_to_alloc, elem)
        if not check_obj:
            continue
        check_rule, alloc = check_allocation_rules(alloc_obj, obj_to_alloc, **kwargs)
        if check_rule and alloc:
            new_allocation[alloc[0]].append(alloc[1])

    update = add_allocation(new_allocation, kwargs['output_xml'])

    return update


def check_allocation_objects_types(alloc_obj, obj_to_alloc, elem):
    """Check that alloc_obj within is FunctionalElement or FunctionalInterface or State or
    PhysicalElement or PhysicalInterface AND obj_to_alloc is Function/State or Data or
    FunctionalElement or FunctionalInterface"""
    check = True
    if alloc_obj is None and obj_to_alloc is None:
        print_wrong_obj_allocation(elem)
        check = False
    elif alloc_obj is None or obj_to_alloc is None:
        if alloc_obj is None:
            print_wrong_obj_allocation(elem[0])
            check = False
        elif obj_to_alloc is None:
            print_wrong_obj_allocation(elem[1])
            check = False

    return check


def print_wrong_obj_allocation(obj_str):
    """Print relative message to wrong pair allocations"""
    if isinstance(obj_str, tuple):
        name_str = f"Objects {obj_str[0]} and {obj_str[1]}"
    else:
        name_str = f"Object {obj_str}"
    print(name_str + " not found or can not be allocated, "
                     "available allocations are: \n"
                     "(Functional Element allocates State/Function) OR \n"
                     "(State allocates Function) OR \n"
                     "(Functional Interface allocates Data) OR \n"
                     "(Physical Element allocates Functional Element) OR \n"
                     "(Physical Interface allocates Functional Interface)\n")


def check_allocation_rules(alloc_obj, obj_to_alloc, **kwargs):
    """Check "good" combinations, trigger specific check and then return check and new tuple
    allocation"""
    check = False
    new_alloc = None
    if isinstance(alloc_obj, datamodel.FunctionalElement):
        if isinstance(obj_to_alloc, (datamodel.Function, datamodel.State)):
            check = True
            pair = check_fun_elem_allocation(alloc_obj, obj_to_alloc, kwargs['xml_fun_elem_list'])
            if pair:
                new_alloc = [0, pair]
        else:
            print_wrong_obj_allocation(obj_to_alloc.name)
    elif isinstance(alloc_obj, datamodel.State):
        if isinstance(obj_to_alloc, datamodel.Function):
            check = True
            pair = check_state_allocation(alloc_obj, obj_to_alloc, kwargs['xml_state_list'])
            if pair:
                new_alloc = [1, pair]
        else:
            print_wrong_obj_allocation(obj_to_alloc.name)
    elif isinstance(alloc_obj, datamodel.FunctionalInterface):
        if isinstance(obj_to_alloc, datamodel.Data):
            check = True
            pair = check_fun_inter_allocation(alloc_obj, obj_to_alloc, **kwargs)
            if pair:
                new_alloc = [2, pair]
        else:
            print_wrong_obj_allocation(obj_to_alloc.name)
    elif isinstance(alloc_obj, datamodel.PhysicalElement):
        if isinstance(obj_to_alloc, datamodel.FunctionalElement):
            check = True
        else:
            print_wrong_obj_allocation(obj_to_alloc.name)
    elif isinstance(alloc_obj, datamodel.PhysicalInterface):
        if isinstance(obj_to_alloc, datamodel.FunctionalInterface):
            check = True
        else:
            print_wrong_obj_allocation(obj_to_alloc.name)

    return check, new_alloc


def check_fun_elem_allocation(fun_elem, obj_to_alloc, fun_elem_list):
    """Check allocation rules for fun_elem then returns objects if check"""
    count = None
    out = None
    check_allocation = get_allocation_object(obj_to_alloc, fun_elem_list)
    if check_allocation is not None:
        count = len(check_allocation)
        for item in check_allocation:
            # Checks if they are in the same family
            if not check_not_family(item, fun_elem) and item != fun_elem:
                count -= 1

    if count in (None, 0):
        if isinstance(obj_to_alloc, datamodel.State):
            fun_elem.add_allocated_state(obj_to_alloc.id)
        else:
            fun_elem.add_allocated_function(obj_to_alloc.id)
        out = [fun_elem, obj_to_alloc]

    return out


def check_state_allocation(state, function, state_list):
    """Check allocation rules for state then returns objects if check"""
    out = None
    check_allocation = get_allocation_object(function, state_list)
    if check_allocation is None:
        state.add_allocated_function(function.id)
        out = [state, function]
    else:
        if state not in check_allocation:
            state.add_allocated_function(function.id)
            out = [state, function]

    return out


def check_fun_inter_allocation(fun_inter, data, **kwargs):
    """Check allocation rules for fun_inter then returns objects if check"""
    out = None
    check_allocation_fun_inter = get_allocation_object(data, kwargs['xml_fun_inter_list'])
    if check_allocation_fun_inter is None:
        check_fe = check_fun_elem_data_consumption(
            data, fun_inter,
            kwargs['xml_fun_elem_list'],
            kwargs['xml_function_list'],
            kwargs['xml_consumer_function_list'],
            kwargs['xml_producer_function_list'])
        if all(i for i in check_fe):
            out = [fun_inter, data]
            fun_inter.add_allocated_data(data.id)
        elif True in check_fe:
            if check_fe[0] is True:
                print(f"Data {data.name} has only consumer(s) "
                      f"allocated to a functional element exposing "
                      f"{fun_inter.name}, {data.name} not "
                      f"allocated to {fun_inter.name}")
            elif check_fe[1] is True:
                print(f"Data {data.name} has only producer(s) "
                      f"allocated to a functional element exposing "
                      f"{fun_inter.name}, {data.name} not "
                      f"allocated to {fun_inter.name}")
        else:
            print(f"Data {data.name} has no producer(s) nor "
                  f"consumer(s) allocated to functional elements "
                  f"exposing {fun_inter.name}, {data.name} not "
                  f"allocated to {fun_inter.name}")
    return out


def check_fun_elem_data_consumption(data, fun_inter, fun_elem_list, function_list,
                                    xml_consumer_function_list, xml_producer_function_list):
    """Check if for a fun_inter, the fun_elem exposing it has allocated functions producing and
    consumming that data"""
    fun_elem_exposes = set()
    for fun_elem in fun_elem_list:
        if any(a == fun_inter.id for a in fun_elem.exposed_interface_list):
            fun_elem_exposes.add(fun_elem)

    cons_list = False
    prod_list = False
    for function in function_list:
        for fun_elem in fun_elem_exposes:
            if any(a == function.id for a in fun_elem.allocated_function_list):
                fun_data = [data.name, function]
                if any(a == fun_data for a in xml_consumer_function_list):
                    cons_list = True
                if any(a == fun_data for a in xml_producer_function_list):
                    prod_list = True

    return [cons_list, prod_list]


def add_allocation(allocation_dict, output_xml):
    """
    Check if allocation_lists is not empty, write in xml for each list and return 0/1
    if some update has been made.

        Parameters:
            allocation_dict : Containing all allocation to write within xml
            output_xml (GenerateXML object) : XML's file object

        Returns:
            1 if update, else 0
    """
    if any(allocation_dict.values()):
        for _, k in enumerate(allocation_dict):
            if allocation_dict[k]:
                output_xml.write_objects_allocation(allocation_dict[k])
                # Warn the user once added within xml
                for elem in allocation_dict[k]:
                    print(f"{elem[1].__class__.__name__} {elem[1].name} is allocated to "
                          f"{elem[0].__class__.__name__} {elem[0].name}")
                    # Check the dict length, if this method is called from viewpoint_orchestrator
                    # or functional_orchestrator for View => Only key[0] and no recursion wanted
                    if k in (0, 1):
                        recursive_allocation(elem, output_xml)
        return 1
    return 0


def check_parent_allocation(elem, output_xml):
    """Check if parent's Function/Sate are allocated to parent's Fucntional Element:
    if not print message to user asking if he wants to, if yes write it in xml then continue
    with parents"""
    if elem[0].parent is not None and elem[1].parent is not None:
        fun_elem_parent = elem[0].parent
        object_parent = elem[1].parent
        check = False
        if isinstance(elem[1], datamodel.State):
            if object_parent.id in fun_elem_parent.allocated_state_list:
                check = True
        elif isinstance(elem[1], datamodel.Function):
            if object_parent.id in fun_elem_parent.allocated_function_list:
                check = True
        if not check:
            answer = input(f"Do you also want to allocate parents(i.e. {object_parent.name} "
                           f"to {fun_elem_parent.name}) ?(Y/N)")
            if answer.lower() == "y":
                if isinstance(elem[1], datamodel.State):
                    fun_elem_parent.add_allocated_state(object_parent.id)
                else:
                    fun_elem_parent.add_allocated_function(object_parent.id)

                add_allocation({5: [[fun_elem_parent, object_parent]]}, output_xml)
                check_parent_allocation([fun_elem_parent, object_parent], output_xml)
            else:
                print(f"Error: {object_parent.name} is not allocated despite at least one "
                      f"of its child is")


def recursive_allocation(elem, output_xml):
    """Recursive allocation for childs of State/Function"""
    check_parent_allocation(elem, output_xml)
    object_type = get_object_type(elem[1])
    if elem[1].child_list:
        for i in elem[1].child_list:
            parent_child = [elem[1], i]
            allocated_child_list = get_allocated_child(parent_child, [elem[0]])
            if allocated_child_list:
                for item in allocated_child_list:
                    if isinstance(elem[1], datamodel.State):
                        item[0].add_allocated_state(item[1].id)
                    else:
                        item[0].add_allocated_function(item[1].id)
                # We want recursivety so it trigger for (0, 1) keys in the dict
                add_allocation({0: allocated_child_list}, output_xml)
    else:
        # TBT/TBC
        if object_type == "state" and elem[1].id not in elem[0].allocated_state_list:
            elem[0].add_allocated_state(elem[1].id)
            print(f"State {elem[1].name} is allocated to functional "
                  f"element {elem[0].name}")
        elif object_type == "function" and elem[1].id not in elem[0].allocated_function_list:
            elem[0].add_allocated_function(elem[1].id)
            print(f"Function {elem[1].name} is allocated to functional "
                  f"element {elem[0].name}")


def get_allocated_child(elem, xml_fun_elem_list):
    """
    Check if the parent state/function is already allocated to a fun elem and create list to add
    its child also (if not already allocated)

        Parameters:
            elem ([State/Function]) : parent object, child object
            xml_fun_elem_list ([FunctionalElement]) : functional element list from xml parsing

        Returns:
            output_list ([FunctionalElement, State/Function]) : Allocation Relationships that need
            to be added
    """
    output_list = []
    for fun_elem in xml_fun_elem_list:
        if isinstance(elem[0], datamodel.State):
            # To avoid "RuntimeError: Set changed size during iteration" copy()
            allocated_list = fun_elem.allocated_state_list.copy()
        else:
            allocated_list = fun_elem.allocated_function_list.copy()
        if allocated_list:
            for allocated_object in allocated_list:
                if allocated_object == elem[0].id:
                    if elem[1].id not in allocated_list:
                        if isinstance(elem[0], datamodel.State):
                            fun_elem.add_allocated_state(elem[1].id)
                        else:
                            fun_elem.add_allocated_function(elem[1].id)
                        output_list.append([fun_elem, elem[1]])

    return output_list


def check_add_inheritance(inherit_str_list, **kwargs):
    """
    Check if each string in allocation_str_list are corresponding to an actual object's name/alias,
    set_derive, create lists of objets.
    Send lists to add_derived() to write them within xml and then returns update from it.

        Parameters:
            inherit_str_list ([str]) : Lists of string from jarvis cell
            kwargs (dict) : whole xml lists + xml's file object

        Returns:
            update ([0/1]) : 1 if update, else 0
    """
    new_list = []
    for elem in inherit_str_list:
        # elem_0: inheriting object
        elem_0 = check_get_object(elem[0],
                                  **{'xml_function_list': kwargs['xml_function_list'],
                                     'xml_fun_elem_list': kwargs['xml_fun_elem_list'],
                                     'xml_fun_inter_list': kwargs['xml_fun_inter_list'],
                                     'xml_phy_elem_list': kwargs['xml_phy_elem_list'],
                                     'xml_phy_inter_list': kwargs['xml_phy_inter_list'],
                                     })
        # elem_1: object to inherit from
        elem_1 = check_get_object(elem[1],
                                  **{'xml_function_list': kwargs['xml_function_list'],
                                     'xml_fun_elem_list': kwargs['xml_fun_elem_list'],
                                     'xml_fun_inter_list': kwargs['xml_fun_inter_list'],
                                     'xml_phy_elem_list': kwargs['xml_phy_elem_list'],
                                     'xml_phy_inter_list': kwargs['xml_phy_inter_list'],
                                     })
        if not elem_0 or not elem_1:
            if not elem_0 and not elem_1:
                print_wrong_object_inheritance(elem[0], elem[1])
            elif not elem_0:
                print_wrong_object_inheritance(elem[0])
            else:
                print_wrong_object_inheritance(elem[1])
            continue
        if elem_0 == elem_1:
            print(f"Same object {elem_0.name}")
            continue
        if elem_0.derived == elem_1:
            continue

        check_obj = check_inheritance(elem_0, elem_1)
        if not check_obj:
            continue
        new_list.append(elem_0)

    if not new_list:
        return 0

    add_derived(new_list, kwargs['output_xml'])
    return 1


def print_wrong_object_inheritance(*obj):
    """Prints wrong object(s) message for inheritance from input string"""
    if len(obj) == 2:
        user_message = f"{obj[0]} and {obj[1]}"
    else:
        user_message = f"{obj[0]}"
    print(f"{user_message} not found, available objects for inheritance are:\n"
          "- Function\n"
          "- Functional element\n"
          "- Functional interface\n"
          "- Physical element\n"
          "- Physical element\n")


def check_inheritance(elem_0, elem_1):
    """Returns check if pair are compatible and object list have been updated"""
    inheritance_type_list = [datamodel.Function, datamodel.FunctionalElement,
                             datamodel.FunctionalInterface, datamodel.PhysicalElement,
                             datamodel.PhysicalInterface]

    type_found = None
    for idx, inheritance_type in enumerate(inheritance_type_list):
        if isinstance(elem_0, inheritance_type) and isinstance(elem_1, inheritance_type):
            type_found = idx
            break
    if type_found is None:
        print(f"{elem_0.__class__.__name__} and {elem_1.__class__.__name__} "
              f"are not of the same type")
        return False
    elem_0.set_derived(elem_1)
    return True


def add_derived(object_list, output_xml):
    """
    Check if input lists are not empty, write in xml for each list and return update if some
    updates has been made.
        Parameters:
            object_list ([Object]) : object with new derived
            output_xml (GenerateXML object) : XML's file object

        Returns:
            1 if update, else 0
    """
    if any(object_list):
        output_xml.write_derived(object_list)
        for obj in object_list:
            print(f"{obj.name} inherited from {obj.derived.name}")
        return 1
    return 0


def check_add_specific_obj_by_type(obj_type_str_list, **kwargs):
    """
    Check if each string in obj_type_str_list are corresponding to an actual object's name/alias,
    set_derive, create object_lists with list per obj. Send lists to add_obj_to_xml()
    write them within xml and then returns update from it.

        Parameters:
            obj_type_str_list ([str]) : Lists of string from jarvis cell
            kwargs (dict) : whole xml lists + xml's file object

        Returns:
            update ([0/1]) : 1 if update, else 0
    """
    object_lists = [[] for _ in range(6)]
    for elem in obj_type_str_list:
        spec_obj_type = None
        if elem[1].capitalize() in [str(i) for i in datamodel.BaseType]:
            base_type = next((i for i in [str(i) for i in datamodel.BaseType]
                              if i == elem[1].capitalize()))
        else:
            spec_obj_type = check_get_object(elem[1],
                                             **{'xml_type_list': kwargs['xml_type_list']})
            base_type = get_base_type_recursively(spec_obj_type)
        if not base_type:
            print(f"No valid base type found for {elem[1]}")
            continue
        base_type = datamodel.BaseType.get_enum(str(base_type)).value
        switch_obj_list = {
            0: create_data_obj,
            1: create_function_obj,
            2: create_fun_elem_obj,
            3: create_fun_inter_obj,
            4: create_phy_elem_obj,
            5: create_phy_inter_obj,
        }
        call = switch_obj_list.get(base_type)
        new_obj = call(elem[0], spec_obj_type, **kwargs)
        if not isinstance(new_obj, int):
            object_lists[base_type].append(new_obj)
        # print(sepc_obj_type.name, sepc_obj_type.base)
    if any(object_lists):
        return add_obj_to_xml(object_lists, kwargs['output_xml'])
    return 0


def get_base_type_recursively(obj_type):
    """Checks type: if it's a BaseType or its base else recursively return """
    if isinstance(obj_type, datamodel.BaseType):
        return obj_type
    elif obj_type.name in [str(i) for i in datamodel.BaseType]:
        return obj_type.name
    elif obj_type.base in [str(i) for i in datamodel.BaseType]:
        return obj_type.base
    return get_base_type_recursively(obj_type.base)


def get_unique_id():
    """Generate and set unique identifier of length 10 integers"""
    identifier = uuid.uuid4()
    return str(identifier.int)[:10]


def add_obj_to_xml(object_lists, output_xml):
    """Send lists object to respective methods of GenerateXML, returns 1 if at least one obj has
    been written"""
    switch_write_obj = {
        0: output_xml.write_data,
        1: output_xml.write_function,
        2: output_xml.write_functional_element,
        3: output_xml.write_functional_interface,
        4: output_xml.write_physical_element,
        5: output_xml.write_physical_interface,
    }
    check = 0
    for idx, sub in enumerate(object_lists):
        if sub:
            switch_write_obj.get(idx)(sub)
            check = 1
    return check


# Inheritance start here - Should be moved/refactored after validation
def childs_inheritance(*xml_obj_set, level=None):
    """Check if object from xml_obj_set is derived, if yes:
    - call derived_childs()
    - returns derived_parent_dict of derived objects id, derived_child_set of derived objects and
    derived_child_id_list
    Parameter:
    - obj: set() of objects
    - level: diagram level"""
    derived_child_set = set()
    derived_parent_dict = {}
    derived_child_id_list = set()

    for xml_set in xml_obj_set:
        for elem in xml_set:
            if elem.derived:
                partial_child_set, partial_par_dict, partial_child_id_set = derived_childs(elem,
                                                                                           level)
                derived_child_set = derived_child_set.union(partial_child_set)
                derived_parent_dict.update(partial_par_dict)
                derived_child_id_list = derived_child_id_list.union(partial_child_id_set)

    return derived_child_set, derived_parent_dict, derived_child_id_list


def derived_childs(obj, level):
    """Add derived object childs to object and set derived parent to object"""
    derived_parent_dict = {}

    [obj.add_child(c) for c in obj.derived.child_list
     if c.parent == obj.derived]
    [c.set_parent(obj) for c in obj.derived.child_list
     if c.parent == obj.derived]
    for child in obj.derived.child_list:
        derived_parent_dict[str(child.id)] = str(obj.id)
    derived_child_set = get_children(obj.derived, level=level)[0]
    derived_child_id_list = {c.id for c in obj.derived.child_list}
    obj.derived.child_list.clear()
    derived_child_set.remove(obj.derived)

    return derived_child_set, derived_parent_dict, derived_child_id_list


def reset_childs_inheritance(*xml_obj_set, derived_child_id=None):
    """Check if there is first level derived child id list, if yes reset original child/parent
    relationships"""
    if derived_child_id:
        for xml_set in xml_obj_set:
            for elem in xml_set:
                if elem.derived:
                    child_pop = [c for c in elem.child_list
                                 if c.id in [f for f in derived_child_id]]
                    [elem.derived.add_child(c) for c in child_pop]
                    [c.set_parent(elem.derived) for c in child_pop]


def attribute_inheritance(xml_attribute_set, *xml_object_set):
    """Attribute inheritance"""
    modified_attrib_set = set()
    derived_elem = set()

    for xml_set in xml_object_set:
        [derived_elem.add(e) for e in xml_set if e.derived]

    for attribute in xml_attribute_set:
        for obj_id, value in attribute.described_item_list.copy():
            for der_obj in derived_elem:
                if obj_id == der_obj.derived.id:
                    attribute.add_described_item((der_obj.id, value))
                    modified_attrib_set.add((attribute.id, der_obj.id, value))

    return modified_attrib_set


def reset_attribute_inheritance(xml_attribute_set, to_be_reset_id_set):
    """Reset attributes"""
    for attribute in xml_attribute_set:
        [attribute.described_item_list.remove((e[1], e[2])) for e in to_be_reset_id_set
         if e[0] == attribute.id]


def view_inheritance(xml_view_set, *xml_obj_set):
    """View inheritance"""
    temp_view_set = set()
    derived_elem_set = set()

    for xml_set in xml_obj_set:
        [derived_elem_set.add(e) for e in xml_set if e.derived]

    for view in xml_view_set:
        for item_id in view.allocated_item_list.copy():
            for obj in derived_elem_set:
                if item_id == obj.derived.id:
                    temp_view_set.add((view.id, obj.id))
                    view.add_allocated_item(obj.id)

    return temp_view_set


def reset_view_inheritance(xml_view_set, to_be_removed_id):
    """Reset views"""
    for view in xml_view_set:
        for ids in to_be_removed_id:
            if view.id == ids[0]:
                view.allocated_item_list.remove(ids[1])


def allocation_inheritance(xml_obj_set, xml_allocated_obj_set):
    """Allocation Inheritance"""
    alloc_to_reset = []
    for elem in xml_obj_set:
        if elem.derived:
            if isinstance(elem, datamodel.FunctionalElement):
                pair = [elem, {i for i in elem.derived.allocated_function_list
                                  if i not in elem.allocated_function_list}]
                elem.allocated_function_list = elem.allocated_function_list.union(
                    elem.derived.allocated_function_list)
                alloc_to_reset.append(pair)

    return alloc_to_reset


def reset_alloc_inheritance(pairs_to_reset):
    """Reset Allocation Inheritance"""
    for pair in pairs_to_reset:
        if isinstance(pair[0], datamodel.FunctionalElement):
            [pair[0].allocated_function_list.remove(fun_id) for fun_id in pair[1]]


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Module datamodel for 3SE"""
from setuptools import setup


setup(name='datamodel',
      version='0.1',
      description='MTC datamodel',
      url='',
      author='',
      author_email='',
      license='',
      packages=['datamodel'],
      zip_safe=False)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Module with class & methods for parsing jarvis4se commands"""
import re
from IPython.display import display, HTML, Markdown

from . import viewpoint_orchestrator
from . import functional_orchestrator
from . import shared_orchestrator
from .question_answer import get_object_list, get_pandas_table, find_question
from .diagram_generator import filter_show_command


class CmdParser:
    def __init__(self, generator):
        self.commands = [
            (r"(under.*)", self.matched_under),

            (r"(?<= |\n)(.*?) extends ([^.|\n]*)", matched_extend),

            (r"(?<= |\n)(.*?) is a ([^state|transition][^\.\n]*)",
             matched_specific_obj),

            (r"(?<= |\n)(.*?) is a state(?=.|\n)", matched_state),

            (r"(?<= |\n)(.*?) is a transition(?=.|\n)", matched_transition),

            (r"(?<= |\n)(.*?) is an attribute\b(?=.|\n)", matched_attribute),

            (r"(?<= |\n)(.*?) inherits from ([^.|\n]*)", matched_inherits),

            (r"The alias of (.*?) is ([^.|\n]*)", matched_alias),

            (r"(?<= |\n)consider ([^.|\n]*)", matched_consider),

            (r"([^\. \.\n]*) is composed of ([^.|\n]*)", matched_composition),

            (r"([^. |.|\n].*) composes ([^.|\n]*)", matched_composition),

            (r"([^. |.|\n].*) consumes ([^.|\n]*)", matched_consumer),

            (r"([^. |.|\n].*) is an input of ([^.|\n]*)", matched_consumer),

            (r"([^. |.|\n].*) produces ([^.|\n]*)", matched_producer),

            (r"([^. |.|\n].*) is an output of ([^.|\n]*)", matched_producer),

            (r"(?<= |\n)(.*?) exposes ([^.|\n]*)", matched_exposes),

            (r"(?<= |\n)(.*?) is allocated to ([^.|\n]*)", matched_allocation),

            (r"(?<= |\n)(.*?) allocates ([^.|\n]*)", matched_allocation),

            (r"(?<= |\n)delete ([^.|\n]*)", matched_delete),

            (r"The type of (.*?) is ([^.|\n]*)", matched_type),

            (r"([^. |.|\n].*) implies ([^.|\n]*)", matched_implies),

            (r"Condition for (.*?) is:([^.|\n]*)", matched_condition),

            (r"The (source|destination) of (.*?) is ([^.|\n]*)", matched_src_dest),

            (r"(?<= |\n)show (.*?)\n", self.matched_show),

            (r"\s([A-Za-z\s].*\?)", matched_question_mark),

            (r"list (input|output|child|data|function|transition|interface) ([^.|\n]*)",
             matched_list),

            (r"The ([^type|alias|source|destination].*?) of (.*?) is ([^.|\n]*)",
             matched_described_attribute),
        ]
        self.reverse = (r"([^. |.|\n].*) composes ([^.|\n]*)",
                        r"([^. |.|\n].*) consumes ([^.|\n]*)",
                        r"([^. |.|\n].*) produces ([^.|\n]*)",
                        r"(?<= |\n)(.*?) is allocated to ([^.|\n]*)")

        self.generator = generator

    def lookup_table(self, string, **kwargs):
        """Lookup table with conditions depending on the match"""
        update_list = []
        for regex, method in self.commands:
            result_chain = None
            result = None
            update = None
            if regex == r"(under.*)":
                result_chain = re.split(regex, string)
                del result_chain[0]
            # Only one diagram per cell can be output
            elif regex == r"(?<= |\n)show (.*?)\n":
                result = re.search(regex, string, re.MULTILINE)
            else:
                # Transform to avoid duplicated function's declaration within cells input
                result = []
                [result.append(x) for x in re.findall(regex, string, re.MULTILINE)
                 if x not in result]

            if result and not result_chain:
                # self.reverse : ("composes", "consumes", "produces", "is allocated to")
                if regex in self.reverse:
                    result = reverse(result)
                update = method(result, **kwargs)

            elif result_chain:
                string = ''
                update = self.matched_under(result_chain, **kwargs)

            if update is not None:
                if isinstance(update, int):
                    update_list.append(update)

        return update_list

    def matched_under(self, chain_name_str, **kwargs):
        """Get "under" declaration"""
        out = []
        for chain, rest in zip(chain_name_str[::2], chain_name_str[1::2]):
            chain = chain.replace("under ", "")
            out.append(viewpoint_orchestrator.add_view(chain,
                                                       kwargs['xml_view_list'],
                                                       kwargs['output_xml']))
            self.lookup_table(rest, **kwargs)
        if 1 in out:
            return 1

        return 0

    def matched_show(self, diagram_name_str, **kwargs):
        """Get "show" declaration"""
        out = filter_show_command(diagram_name_str, **kwargs)
        if out:
            url = self.generator.get_diagram_url(out)
            hyper = get_hyperlink(url)
            display(HTML(hyper))
            print("Overview :")
            display(Markdown(f'![figure]({url})'))


def matched_extend(type_str_list, **kwargs):
    """Get extend declaration"""
    out = viewpoint_orchestrator.check_set_extends(type_str_list,
                                                   kwargs['xml_type_list'],
                                                   kwargs['output_xml'])
    return out


def matched_specific_obj(obj_type_str, **kwargs):
    """Get "is a" declaration"""
    out = shared_orchestrator.check_add_specific_obj_by_type(obj_type_str, **kwargs)
    return out


def matched_state(state_name_str_list, **kwargs):
    """Get state's declaration"""
    out = functional_orchestrator.add_state_by_name(state_name_str_list,
                                                    kwargs['xml_state_list'],
                                                    kwargs['output_xml'])
    return out


def matched_transition(transition_name_str_list, **kwargs):
    """Get transition's declaration"""
    out = functional_orchestrator.add_transition_by_name(transition_name_str_list,
                                                         kwargs['xml_transition_list'],
                                                         kwargs['output_xml'])
    return out


def matched_attribute(attribute_name_str, **kwargs):
    """Get "attribute" declaration"""
    out = viewpoint_orchestrator.add_attribute(attribute_name_str,
                                               kwargs['xml_attribute_list'],
                                               kwargs['output_xml'])
    return out


def matched_inherits(inherits_str, **kwargs):
    """Get inherits from declaration"""
    out = shared_orchestrator.check_add_inheritance(inherits_str, **kwargs)
    return out


def matched_alias(alias_str_list, **kwargs):
    """Get "alias" declaration"""
    out = shared_orchestrator.check_set_object_alias(alias_str_list, **kwargs)
    return out


def matched_consider(consider_str_list, **kwargs):
    """Get "consider" declaration"""
    out = viewpoint_orchestrator.check_get_consider(consider_str_list,
                                                    kwargs['xml_function_list'],
                                                    kwargs['xml_fun_elem_list'],
                                                    kwargs['xml_data_list'],
                                                    kwargs['xml_view_list'],
                                                    kwargs['output_xml'])
    return out


def matched_composition(parent_child_name_str_list, **kwargs):
    """Get composition relationship command (match for "is composed by' or "composes")"""
    out = shared_orchestrator.check_add_child(parent_child_name_str_list,
                                              **{
                                                  'xml_function_list': kwargs['xml_function_list'],
                                                  'xml_state_list': kwargs['xml_state_list'],
                                                  'xml_fun_elem_list': kwargs['xml_fun_elem_list'],
                                                  'xml_phy_elem_list': kwargs['xml_phy_elem_list'],
                                                  'output_xml': kwargs['output_xml'],
                                              })
    return out


def matched_consumer(consumer_str_list, **kwargs):
    """Get consumer declaration"""
    out = functional_orchestrator.check_add_consumer_function(
        consumer_str_list,
        kwargs['xml_consumer_function_list'],
        kwargs['xml_producer_function_list'],
        kwargs['xml_function_list'],
        kwargs['xml_data_list'],
        kwargs['output_xml'])
    return out


def matched_producer(producer_str_list, **kwargs):
    """Get producer declaration"""
    out = functional_orchestrator.check_add_producer_function(
        producer_str_list,
        kwargs['xml_consumer_function_list'],
        kwargs['xml_producer_function_list'],
        kwargs['xml_function_list'],
        kwargs['xml_data_list'],
        kwargs['output_xml'])
    return out


def matched_allocation(allocation_str_list, **kwargs):
    """Get allocation declaration"""
    out = shared_orchestrator.check_add_allocation(allocation_str_list, **kwargs)
    return out


def matched_exposes(exposes_str_list, **kwargs):
    """Get 'exposes' declaration"""
    out = functional_orchestrator.check_add_exposes(exposes_str_list,
                                                    kwargs['xml_fun_elem_list'],
                                                    kwargs['xml_fun_inter_list'],
                                                    kwargs['xml_data_list'],
                                                    kwargs['output_xml'])
    return out


def matched_delete(delete_str_list, **kwargs):
    """Get delete declaration"""
    out = shared_orchestrator.check_and_delete_object(delete_str_list, **kwargs)
    return out


def matched_type(type_str_list, **kwargs):
    """Get set_type declaration"""
    out = shared_orchestrator.check_set_object_type(type_str_list, **kwargs)
    return out


def matched_implies(data_predecessor_str_set, **kwargs):
    """Get predecessor declaration"""
    out = functional_orchestrator.check_add_predecessor(data_predecessor_str_set,
                                                        kwargs['xml_data_list'],
                                                        kwargs['xml_view_list'],
                                                        kwargs['output_xml'])
    return out


def matched_condition(condition_str_list, **kwargs):
    """Get set_condition declaration"""
    out = functional_orchestrator.check_add_transition_condition(condition_str_list,
                                                                 kwargs['xml_transition_list'],
                                                                 kwargs['output_xml'])
    return out


def matched_src_dest(src_dest_str, **kwargs):
    """Get source/destination declaration for transition"""
    out = functional_orchestrator.check_add_src_dest(src_dest_str,
                                                     kwargs['xml_transition_list'],
                                                     kwargs['xml_state_list'],
                                                     kwargs['output_xml'])
    return out


def get_hyperlink(path):
    """Convert file path into clickable form."""
    text = "Click to open in new tab"
    # convert the url into link
    return f'<a href="{path}" target="_blank">{text}</a>'


def matched_question_mark(question_str, **kwargs):
    """Gets "?" declaration"""
    out = find_question(question_str, **kwargs)
    if out:
        for elem in out:
            if isinstance(elem, str):
                print(elem)
            else:
                display(elem)


def matched_list(object_str, **kwargs):
    """Gets list declaration"""
    out = get_object_list(object_str, **kwargs)
    if out:
        for i in out:
            display(HTML(get_pandas_table(i)))


def matched_described_attribute(described_attribute_str, **kwargs):
    """Get the described attribute value for an object"""
    out = viewpoint_orchestrator.check_add_object_attribute(described_attribute_str,
                                                            kwargs['xml_attribute_list'],
                                                            kwargs['xml_function_list'],
                                                            kwargs['xml_fun_elem_list'],
                                                            kwargs['xml_fun_inter_list'],
                                                            kwargs['output_xml'])
    return out


def reverse(inverted_list):
    """Reverses input tuple strings"""
    sorted_list = []
    for tuples in inverted_list:
        new_tup = ()
        for k in reversed(tuples):
            new_tup = new_tup + (k,)
        sorted_list.append(new_tup)
    return sorted_list



#!/usr/bin/env python
"""Modules for MagicTools class i.e. magic line %retrieve_pkg_version and magic cell %%diagram"""
# Libraries
from sys import version as py_ver
from importlib.metadata import version
from IPython.core.magic import (Magics, magics_class, cell_magic, line_magic)
from IPython.display import display, HTML, Markdown

# Modules
from jarvis.command_parser import get_hyperlink


@magics_class
class MagicTools(Magics):
    """Magic class"""
    def __init__(self, shell, generator):
        # You must call the parent constructor
        super().__init__(shell)
        self.generator = generator

    @staticmethod
    @line_magic
    def retrieve_pkg_version(_):
        """Magic line %retrieve_pkg_version that get dependecies versions, for users to share when
        they create issues"""
        pkg = ['ipython', 'lxml', 'notebook', 'plantuml', 'jarvis4se', 'pandas']
        pkg_ver = "\n".join(['=='.join(tups) for tups in list(zip(pkg, list(map(version, pkg))))])
        print(pkg_ver, f"\npython=={py_ver[:6]}")

    @cell_magic
    def diagram(self, _, cell):
        """Magic cell %%diagram """
        out = self.generator.get_diagram_url(cell, True)
        if out:
            hyper = get_hyperlink(out)
            display(HTML(hyper))
            print("Overview :")
            display(Markdown(f'![figure]({out})'))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Module extension for datamodel"""


class Point:
    """Point class: Compatible with LTPSICE"""
    def __init__(self, p_x='', p_y=''):
        """Init"""
        self.x = p_x
        self.y = p_y

    def set_x(self, p_x):
        """Set x"""
        self.x = p_x

    def set_y(self, p_y):
        """Set y"""
        self.y = p_y


class EndPoint:
    """End point class: Compatible with LTSPICE"""
    def __init__(self, p_name=''):
        """Init"""
        self.name = p_name
        self.point_list = set()

    def set_name(self, p_name):
        """Set name"""
        self.name = p_name

    def add_point(self, p_point):
        """Add point to point_list"""
        self.point_list.add(p_point)



from .plantuml_adapter import get_function_diagrams
from .plantuml_adapter import get_sequence_diagram
from .plantuml_adapter import get_state_machine_diagram
from .plantuml_adapter import get_fun_elem_decomposition
from .plantuml_adapter import get_fun_elem_context_diagram
from .util import PlantUmlPicoServer
from .util import PlantUmlGen
